Package: statcanR
Type: Package
Title: Client for Statistics Canada's Open Economic Data
Version: 0.3.9
Authors@R: c(
    person("Thierry", "Warin", email = "thierry.warin@hec.ca",
           role = c("aut", "cre"),
           comment = c(ORCID = "0000-0002-5921-3428")))
Description: Provides an R client for Statistics Canada's Web Data Service.
    Users can describe the data they need in natural language, search the
    official table catalogue, and download complete data tables in English or
    French as data frames. Tables formerly known as CANSIM tables are
    identified by Product IDs. Warin (2024)
    <doi:10.5070/T5.1868>.
Depends: R (>= 4.0.0)
License: MIT + file LICENSE
Encoding: UTF-8
Language: en-US
Suggests: knitr, rmarkdown, testthat (>= 3.0.0), withr
VignetteBuilder: knitr
Imports: data.table, jsonlite, DT, httr
URL: https://warint.github.io/statcanR/,
        https://github.com/warint/statcanR
BugReports: https://github.com/warint/statcanR/issues
Config/testthat/edition: 3
Config/Needs/website: pkgdown
Config/roxygen2/version: 8.0.0
NeedsCompilation: no
Packaged: 2026-07-25 04:54:49 UTC; warint
Author: Thierry Warin [aut, cre] (ORCID:
    <https://orcid.org/0000-0002-5921-3428>)
Maintainer: Thierry Warin <thierry.warin@hec.ca>
Repository: CRAN
Date/Publication: 2026-07-25 05:10:02 UTC

# Generated by roxygen2: do not edit by hand

S3method(print,statcan_chat_result)
export(statcan_chat)
export(statcan_chat_continue)
export(statcan_data)
export(statcan_download_data)
export(statcan_find)
export(statcan_search)

<!-- README.md is generated from README.Rmd. Please edit README.Rmd. -->

# statcanR

<!-- badges: start -->

[![R-CMD-check](https://github.com/warint/statcanR/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/warint/statcanR/actions/workflows/R-CMD-check.yaml)
[![CRAN
status](https://www.r-pkg.org/badges/version/statcanR)](https://CRAN.R-project.org/package=statcanR)
[![CRAN
downloads](https://cranlogs.r-pkg.org/badges/grand-total/statcanR?color=blue)](https://cran.r-project.org/package=statcanR)
[![Mentioned in
Awesome](https://awesome.re/mentioned-badge.svg)](https://github.com/SNStatComp/awesome-official-statistics-software)
<!-- badges: end -->

`statcanR` helps you find and download data tables published by
Statistics Canada. It connects to the official [Web Data Service
(WDS)](https://www.statcan.gc.ca/en/developers/wds), works in English or
French, and returns ordinary data frames that can be analysed with base
R or your preferred R packages.

The package has four main functions, plus one optional one:

| If you want to…                                                                                        | Use…                      | What you get                                                           |
|--------------------------------------------------------------------------------------------------------|---------------------------|------------------------------------------------------------------------|
| Describe the data you need in ordinary language                                                        | `statcan_find()`          | Ranked table choices, identifiers, and an explanation of each match    |
| Search for exact words in table titles                                                                 | `statcan_search()`        | An interactive table of matching titles and identifiers                |
| Load a complete table into R                                                                           | `statcan_data()`          | A data frame                                                           |
| Load a table and also save a CSV copy                                                                  | `statcan_download_data()` | A data frame and a UTF-8 CSV file                                      |
| Get a language model’s help interpreting a `statcan_find()` query *(optional, requires configuration)* | `statcan_chat()`          | An explanation of the best match and a clarifying question when needed |

`statcanR` downloads the **complete** Statistics Canada table. Some
tables are large, so check that the table is appropriate for your needs
before downloading it.

## Installation and upgrades

Install or upgrade the released package with the same command:

``` r
install.packages("statcanR")
```

To install the development version from GitHub:

``` r
install.packages("remotes")
remotes::install_github("warint/statcanR")
```

If you upgrade while `statcanR` is already loaded, restart your R
session before loading it again. You can confirm the installed version
with:

``` r
packageVersion("statcanR")
```

Starting with version 0.3.x, the established calls to
`statcan_search()`, `statcan_data()`, and `statcan_download_data()` are
unchanged, so scripts written for earlier releases continue to work. The
new `statcan_find()` function adds a more conversational way to discover
a table without changing those functions.

## A first workflow

Start by loading the package:

``` r
library(statcanR)
```

### 1. Find a table

Describe the subject, place, and period you need when you do not yet
know the table identifier:

``` r
matches <- statcan_find(
  "R&D expenditures in Quebec since 2020",
  lang = "eng",
  n = 5
)

matches[, c("title", "id", "score", "match_reason")]
```

`statcan_find()` interprets this request as three clues:

1.  **Subject:** research and development expenditures;
2.  **Geography:** Quebec; and
3.  **Coverage:** a table containing data for 2020.

It returns an ordinary data frame, ranked from the strongest match to
the weakest. The `id` column contains the identifier needed by the
download functions. The `match_reason` column explains why each table
was included, so read the titles before selecting one. Several tables
can answer different interpretations of the same request.

The geography and date are used to check the **table as a whole**. They
do not filter the observations that will later be downloaded. After
downloading, select Quebec and the years from 2020 onward using the
relevant columns in that particular table.

If you already know the exact words used in a Statistics Canada title,
use `statcan_search()` instead:

``` r
statcan_search(
  c("federal", "expenditures", "objectives"),
  lang = "eng"
)
```

Searches are case-insensitive. When you supply several keywords,
**every** keyword must appear in the title. If a search is too narrow,
try fewer or more general words. To search French titles, use
`lang = "fra"`.

### Optional: ask a language model for help

`statcan_chat()` wraps `statcan_find()` with a language model you
configure. It explains which candidate best matches your query and asks
a clarifying question when the query is ambiguous — it never invents a
table number or reasons over the downloaded data itself, so
`statcan_find()` remains the authoritative source of candidates.

#### Who is this function for?

`statcan_chat()` is **not** required to use `statcanR`. Finding and
downloading data — `statcan_find()`, `statcan_search()`,
`statcan_data()` — works with no account, no key, and no configuration.
`statcan_chat()` is a convenience layer on top of `statcan_find()` for
people who prefer to *describe* the data they want in plain language and
get a short explanation of the shortlist, rather than skim the ranked
titles themselves. Think of it as an optional research assistant, useful
when:

- you are new to the Statistics Canada catalogue and don’t yet know
  which table numbers to look for;
- a query is ambiguous and you’d like a nudge (a clarifying question)
  before committing;
- you’re teaching or demonstrating and want the reasoning spelled out.

Because it talks to a language model, it needs *one* of three things, in
decreasing order of setup effort and cost:

1.  an **API key** from a commercial provider (OpenAI or
    Anthropic/Claude) — pay-as-you-go, billed separately from any
    ChatGPT Plus or Claude Pro subscription;
2.  a **free-tier key** from an OpenAI-compatible provider (for example
    Groq or Google Gemini);
3.  **nothing at all** — a model you run yourself, locally, on your own
    computer (see [No API key? Run a model on your own
    computer](#no-api-key-run-a-model-on-your-own-computer) below).

> **A subscription is not API access.** A ChatGPT Plus or Claude Pro
> subscription lets you chat on the provider’s website; it does *not*
> give you a programmatic key that R can use. There is no supported “log
> in through the website” flow for a package like this. If you don’t
> want to create a billed API account, use option 2 or 3.

This is entirely optional: it adds no new package dependencies, and it
makes no network request unless you call `statcan_chat()` yourself.
Choose a provider with the `provider` argument — `"openai"` (the
default) or `"anthropic"` (Claude). The endpoint must use `https://`
(plain `http://` is accepted only for a loopback host such as
`http://localhost`, for a local model), so the key is never sent in
cleartext. Set your provider’s API key once per session (the endpoint
follows from the provider; you pass the `model` you want):

``` r
# The API key is a secret, so it is read from an environment variable rather
# than options() (which can be dumped, saved with a session, or land in
# .Rhistory).
Sys.setenv(OPENAI_API_KEY = "sk-...")        # OpenAI
Sys.setenv(ANTHROPIC_API_KEY = "sk-ant-...") # Anthropic (Claude)
```

`library(statcanR)` never asks for these settings, and no key is needed
for `statcan_find()` or `statcan_data()` — they are read only when you
call `statcan_chat()`. To set the key once and reuse it in every session
without retyping it, add it to your `~/.Renviron` file (open it with
`usethis::edit_r_environ()`, then restart R) rather than calling
`Sys.setenv()` each time:

    OPENAI_API_KEY=sk-...
    ANTHROPIC_API_KEY=sk-ant-...

R reads `.Renviron` automatically at startup, so the key stays out of
your scripts and `.Rhistory`.

``` r
# OpenAI (the default provider)
statcan_chat("R&D expenditures in Quebec since 2020", model = "gpt-4o-mini")

# Anthropic (Claude)
statcan_chat(
  "R&D expenditures in Quebec since 2020",
  provider = "anthropic", model = "claude-opus-4-8"
)
```

#### No API key? Run a model on your own computer

If you don’t have — and don’t want to pay for — an API key, you can run
an open-source language model **locally**. The model lives on your own
machine, so there is no account to create, nothing to pay, and your
queries never leave your computer. This is often the best choice for
teaching, for privacy-sensitive work, or simply for trying the feature
out.

The most beginner-friendly option is [**Ollama**](https://ollama.com):
install it, then in a terminal run `ollama pull llama3.1` once to
download a model and `ollama serve` to start it. Ollama exposes an
OpenAI-compatible endpoint at `http://localhost:11434`, so
`statcan_chat()` reaches it through the default `"openai"` provider —
you just point `endpoint` at your local server:

``` r
statcan_chat(
  "R&D expenditures in Quebec since 2020",
  endpoint = "http://localhost:11434/v1/chat/completions",
  api_key = "ollama",   # a placeholder; a local server ignores the key
  model = "llama3.1"    # any model you have pulled with `ollama pull`
)
```

A few things worth knowing:

- **`http://localhost` is allowed on purpose.** Normally
  `statcan_chat()` refuses plain `http://` so your key can never travel
  unencrypted, but a loopback address (`localhost`) never leaves your
  machine, so it is safe and permitted.
- **The `api_key` is a placeholder.** Local servers don’t check it, but
  the argument is still required, so pass any non-empty string
  (`"ollama"` is a convention).
- **`model` must be one you have downloaded** — whatever you pulled with
  `ollama pull` (for example `"llama3.1"`, `"mistral"`, `"qwen2.5"`).

The same pattern works for [LM Studio](https://lmstudio.ai) and other
OpenAI-compatible local servers, as well as free hosted options like
Groq or Google Gemini — point `endpoint` (and, for hosted ones,
`api_key`) at the service and keep `provider = "openai"`.

#### What `statcan_chat()` gives back

`statcan_chat()` does **not** narrow your query down to a single table.
Like `statcan_find()`, it returns *several* ranked candidates — the
language model only explains them; it never picks or invents one for
you. When you print the result you see the whole shortlist:

``` r
Query: R&D expenditures in Quebec since 2020

  rank                                                 title       id
1    1 Gross domestic expenditures on research and develo... 27100359
2    2 Personnel engaged in research and development (R&D...  27100341
3    3 ...                                                    ...

<the model's explanation of which candidate fits best>

Clarifying question: Do you want spending totals, or the number of personnel?
```

The value it returns is a list with four named parts, so you can reach
any piece of it directly instead of copying identifiers by hand:

``` r
result <- statcan_chat("R&D expenditures in Quebec since 2020")

result$query               # the query you asked, unchanged
result$candidates          # the full data frame from statcan_find()
result$explanation         # the model's plain-language explanation
result$clarifying_question # a follow-up question, or NA when the query was clear
```

The important part is `result$candidates`: it is the **same data frame**
that `statcan_find()` produces, already containing every identifier and
its ranking. You therefore never need to retype a table number — just
index into that data frame:

``` r
result$candidates$id       # every candidate identifier, best-ranked first
result$candidates$id[1]    # just the top-ranked identifier
```

That last value is exactly what `statcan_data()` expects, so a full
workflow — describe, choose, download — becomes three short lines with
nothing copied by hand:

``` r
result     <- statcan_chat("R&D expenditures in Quebec since 2020")
top_id     <- result$candidates$id[1]
table_data <- statcan_data(top_id, lang = "eng")
```

If more than one candidate looks plausible, read `result$explanation`
and the `title` column of `result$candidates`, then pick the row you
want (for example `result$candidates$id[2]`) before calling
`statcan_data()`.

#### Continue the conversation

A single `statcan_chat()` call is one turn. If it comes back with a
`clarifying_question`, you can answer it — and keep the discussion going
— with `statcan_chat_continue()`. Each follow-up stays scoped to the
**same candidate tables** the first call already found: it does not run
a new catalogue search, and the model still never invents a table
number. Results are chainable, so you can answer more than once. Just
reassign the same variable each turn — there is no need for
`r1`/`r2`/`r3`-style names:

``` r
chat <- statcan_chat("R&D spending in Quebec", provider = "anthropic", model = "claude-opus-4-8")
chat$clarifying_question           # e.g. "Annual or quarterly data?"

chat <- statcan_chat_continue(chat, "annual data, since 2015")
chat$explanation                   # the model's updated take, same shortlist

chat <- statcan_chat_continue(chat, "just the total, not by industry")
```

To search the catalogue again from scratch, start a fresh
`statcan_chat()` call rather than continuing. The API key is re-read on
each `statcan_chat_continue()` call (from the argument or your
environment variable) and is never stored inside the result object.

### 2. Download the table into R

Copy an identifier from the search result and pass it to
`statcan_data()`. The example below uses a small table so it is
convenient to try:

``` r
table_data <- statcan_data("10-10-0001-01", lang = "eng")
```

The result is a data frame. Inspect its size, column names, and first
rows before beginning an analysis:

``` r
dim(table_data)
names(table_data)
head(table_data)
```

`REF_DATE` contains the reference period and is converted to a `Date`
when the source format can be interpreted safely. Coordinate columns
remain character values, and `INDICATOR` contains the official table
title.

### 3. Understand table identifiers and languages

Statistics Canada displays identifiers such as `10-10-0001-01`. The
corresponding eight-digit Product ID (PID) is `10100001`. `statcanR`
accepts either form, so these calls request the same table:

``` r
table_data <- statcan_data("10-10-0001-01", "eng")
table_data <- statcan_data("10100001", "eng")
```

Use `lang = "eng"` for English or `lang = "fra"` for French:

``` r
table_fr <- statcan_data("10-10-0001-01", "fra")
```

Column labels supplied by Statistics Canada may differ between the
English and French tables.

## Save a CSV file

Use `statcan_data()` when you only need the data in R. Use
`statcan_download_data()` when you also want a CSV copy. Earlier
two-argument calls remain valid and save into the current working
directory:

``` r
table_data <- statcan_download_data("10-10-0001-01", "eng")
getwd()
```

For clearer file management, create an output directory and provide it
with `path`:

``` r
output_dir <- file.path(tempdir(), "statcanR-data")
dir.create(output_dir, showWarnings = FALSE)

table_data <- statcan_download_data(
  "10-10-0001-01",
  "eng",
  path = output_dir
)

attr(table_data, "statcan_file")
```

The function still returns the data frame. The `statcan_file` attribute
records the exact path of the saved CSV file.

## Catalogue and metadata caching

The table catalogue is cached for 24 hours so repeated searches are
fast. `statcan_find()` also caches the candidate metadata used to verify
a geography for seven days. Use `refresh = TRUE` only when you need the
newest catalogue and metadata:

``` r
statcan_find(
  "population in Alberta since 2021",
  lang = "eng",
  refresh = TRUE
)
```

Downloading and refreshing require an internet connection. If Statistics
Canada’s service is temporarily unavailable, searches can use an
existing valid cache. A natural-language search can still return
candidates when geography metadata is unavailable; in that case,
`geography_match` is `NA` and the explanation says that the geography
could not be verified. Confirm the table title, identifier, language,
and output directory before retrying.

For the complete walkthrough, open the installed vignette:

``` r
vignette("getting-started", package = "statcanR")
```

## Licence and citation

Statistics Canada data are provided under the [Statistics Canada Open
Licence](https://www.statcan.gc.ca/en/terms-conditions/open-licence).
The `statcanR` package is released under the MIT licence.

To cite the package and its methodology, run:

``` r
citation("statcanR")
```

The preferred methodological reference is:

> Warin, T. (2024). Access Statistics Canada’s Open Economic Data for
> Statistics and Data Science Courses. *Technology Innovations in
> Statistics Education*, 15(1). <https://doi.org/10.5070/T5.1868>

## Acknowledgements

The author thanks the Center for Interuniversity Research and Analysis
of Organizations (CIRANO) for its support, along with Thibault Senegas,
Jeremy Schneider, Marine Leroi, Martin Paquette, and contributors to
earlier versions of the package. Errors and omissions remain the
author’s.

#' Search Statistics Canada data tables
#'
#' Searches the catalogue of tables published through Statistics Canada's Web
#' Data Service (WDS). The catalogue is cached for 24 hours in the user's cache
#' directory. If WDS is temporarily unavailable, the most recent valid cache is
#' used.
#'
#' @param keywords Character vector of words that must appear in the table
#'   title. Matching is case-insensitive.
#' @param lang Language of the returned titles: `"eng"` or `"fra"`.
#' @param refresh Logical; if `TRUE`, ignore a fresh cache and request the
#'   catalogue from WDS.
#'
#' @return A `DT::datatable` containing matching table titles, Product IDs,
#'   release dates, and language.
#' @export
#'
#' @examples
#' \dontrun{
#' statcan_search(c("economy", "export"), "eng")
#' statcan_search("population", "fra")
#' }
statcan_search <- function(keywords, lang = c("eng", "fra"),
                           refresh = FALSE) {
  lang <- match.arg(lang)

  if (!is.character(keywords) || !length(keywords) || anyNA(keywords)) {
    stop("`keywords` must be a non-empty character vector.", call. = FALSE)
  }
  keywords <- trimws(keywords)
  if (any(!nzchar(keywords))) {
    stop("`keywords` must not contain empty strings.", call. = FALSE)
  }
  if (!is.logical(refresh) || length(refresh) != 1L || is.na(refresh)) {
    stop("`refresh` must be TRUE or FALSE.", call. = FALSE)
  }

  results <- filter_statcan_catalogue(
    statcan_catalogue(refresh = refresh),
    keywords,
    lang
  )

  DT::datatable(
    results,
    rownames = FALSE,
    options = list(pageLength = 10L, searchHighlight = TRUE)
  )
}


# Retrieve and cache the WDS table catalogue.
statcan_catalogue <- function(refresh = FALSE) {
  cache_dir <- tools::R_user_dir("statcanR", which = "cache")
  cache_file <- file.path(cache_dir, "statcan_catalogue.rds")
  cache_ttl <- 24 * 60 * 60

  cache_is_fresh <- file.exists(cache_file) &&
    as.numeric(difftime(Sys.time(), file.info(cache_file)$mtime,
                        units = "secs")) < cache_ttl

  if (!refresh && cache_is_fresh) {
    cached <- read_cached_catalogue(cache_file)
    if (is_valid_statcan_catalogue(cached)) {
      return(cached)
    }
  }

  endpoint <- paste0(
    "https://www150.statcan.gc.ca/",
    "t1/wds/rest/getAllCubesListLite"
  )
  downloaded <- tryCatch({
    response <- httr::GET(
      endpoint,
      httr::timeout(60),
      httr::user_agent(statcan_user_agent())
    )
    httr::stop_for_status(response)

    payload <- jsonlite::fromJSON(
      httr::content(response, as = "text", encoding = "UTF-8"),
      simplifyDataFrame = TRUE
    )
    normalize_statcan_catalogue(payload)
  }, error = function(error) {
    if (file.exists(cache_file)) {
      cached <- read_cached_catalogue(cache_file)
      if (is_valid_statcan_catalogue(cached)) {
        warning(
          "Statistics Canada WDS is unavailable; using the cached catalogue. ",
          conditionMessage(error),
          call. = FALSE
        )
        return(cached)
      }
    }

    stop(
      "Unable to retrieve the Statistics Canada table catalogue and no ",
      "valid cache is available. ", conditionMessage(error),
      call. = FALSE
    )
  })

  dir.create(cache_dir, recursive = TRUE, showWarnings = FALSE)
  tryCatch(
    saveRDS(downloaded, cache_file, version = 3L),
    error = function(error) {
      warning("Unable to cache the Statistics Canada catalogue: ",
              conditionMessage(error), call. = FALSE)
    }
  )
  downloaded
}


read_cached_catalogue <- function(cache_file) {
  catalogue <- tryCatch(readRDS(cache_file), error = function(error) NULL)
  if (is.data.frame(catalogue)) {
    title_columns <- intersect(c("title_eng", "title_fra"), names(catalogue))
    for (title_column in title_columns) {
      if (is.character(catalogue[[title_column]])) {
        catalogue[[title_column]] <- decode_statcan_entities(
          catalogue[[title_column]]
        )
      }
    }
  }
  catalogue
}


# Decode the XML entities used in some WDS catalogue titles. Ampersands are
# decoded last so text that was deliberately escaped twice is only decoded by
# one layer.
decode_statcan_entities <- function(text) {
  result <- as.character(text)
  # Every entity we decode begins with "&", so titles without one need no work.
  # Almost no catalogue titles contain entities, so restricting the fixed-string
  # substitutions to that subset avoids scanning tens of thousands of strings on
  # every cache read.
  candidates <- grepl("&", result, fixed = TRUE)
  if (!any(candidates)) {
    return(result)
  }

  replacements <- c(
    "&quot;" = "\"",
    "&#34;" = "\"",
    "&apos;" = "'",
    "&#39;" = "'",
    "&#039;" = "'",
    "&lt;" = "<",
    "&gt;" = ">",
    "&nbsp;" = "\u00a0",
    "&amp;" = "&"
  )
  subset <- result[candidates]
  for (entity in names(replacements)) {
    subset <- gsub(entity, replacements[[entity]], subset, fixed = TRUE)
  }
  result[candidates] <- subset
  result
}


# Convert the WDS response to a stable, package-owned schema.
normalize_statcan_catalogue <- function(payload) {
  if (is.list(payload) && !is.data.frame(payload) &&
      identical(payload$status, "SUCCESS") && !is.null(payload$object)) {
    payload <- payload$object
  }
  if (!is.data.frame(payload)) {
    stop("Statistics Canada WDS returned an unexpected catalogue format.",
         call. = FALSE)
  }

  required <- c("productId", "cubeTitleEn", "cubeTitleFr", "releaseTime")
  missing <- setdiff(required, names(payload))
  if (length(missing)) {
    stop(
      "Unexpected response from Statistics Canada WDS; missing field(s): ",
      paste(missing, collapse = ", "),
      call. = FALSE
    )
  }

  product_id <- format_product_id_vector(payload$productId)
  release_date <- suppressWarnings(
    as.Date(substr(as.character(payload$releaseTime), 1L, 10L))
  )
  start_date <- catalogue_date_field(payload, "cubeStartDate")
  end_date <- catalogue_date_field(payload, "cubeEndDate")
  archived <- if ("archived" %in% names(payload)) {
    normalize_archive_status(payload$archived)
  } else {
    rep(NA, nrow(payload))
  }
  frequency_code <- if ("frequencyCode" %in% names(payload)) {
    suppressWarnings(as.integer(payload$frequencyCode))
  } else {
    rep(NA_integer_, nrow(payload))
  }
  catalogue <- data.frame(
    title_eng = decode_statcan_entities(payload$cubeTitleEn),
    title_fra = decode_statcan_entities(payload$cubeTitleFr),
    id = product_id,
    start_date = start_date,
    end_date = end_date,
    release_date = release_date,
    archived = archived,
    frequency_code = frequency_code,
    stringsAsFactors = FALSE
  )

  if (!is_valid_statcan_catalogue(catalogue)) {
    stop("Statistics Canada WDS returned an invalid catalogue.", call. = FALSE)
  }

  catalogue
}


catalogue_date_field <- function(payload, field) {
  if (!field %in% names(payload)) {
    return(rep(as.Date(NA), nrow(payload)))
  }

  value <- as.character(payload[[field]])
  value[!nzchar(value)] <- NA_character_
  suppressWarnings(as.Date(substr(value, 1L, 10L)))
}


# WDS currently uses code 2 for a current table and code 1 for an archive.
# Logical values are also accepted because the field is documented as Boolean.
normalize_archive_status <- function(value) {
  if (is.logical(value)) {
    return(value)
  }

  normalized <- tolower(trimws(as.character(value)))
  result <- rep(NA, length(normalized))
  result[normalized %in% c("1", "true", "archived")] <- TRUE
  result[normalized %in% c("0", "2", "false", "current")] <- FALSE
  result
}


format_product_id_vector <- function(product_id) {
  product_id <- if (is.numeric(product_id)) {
    sprintf("%08.0f", product_id)
  } else {
    gsub("[^0-9]", "", as.character(product_id))
  }
  product_id <- ifelse(
    nchar(product_id) == 10L,
    substr(product_id, 1L, 8L),
    product_id
  )

  valid <- grepl("^[0-9]{8}$", product_id)
  formatted <- rep(NA_character_, length(product_id))
  formatted[valid] <- paste0(
    substr(product_id[valid], 1L, 2L), "-",
    substr(product_id[valid], 3L, 4L), "-",
    substr(product_id[valid], 5L, 8L), "-01"
  )
  formatted
}


# Filter a normalized catalogue without making a network request.
filter_statcan_catalogue <- function(catalogue, keywords, lang) {
  title_column <- if (lang == "eng") "title_eng" else "title_fra"
  searchable <- tolower(catalogue[[title_column]])
  keywords <- unique(tolower(keywords))

  matches <- Reduce(`&`, lapply(keywords, function(keyword) {
    grepl(keyword, searchable, fixed = TRUE)
  }))
  number_of_matches <- sum(matches)

  data.frame(
    title = catalogue[[title_column]][matches],
    id = catalogue$id[matches],
    description = rep(NA_character_, number_of_matches),
    release_date = catalogue$release_date[matches],
    lang = rep(lang, number_of_matches),
    stringsAsFactors = FALSE
  )
}


# Reject corrupt caches and caches created with an incompatible schema.
is_valid_statcan_catalogue <- function(catalogue) {
  required <- c(
    "title_eng", "title_fra", "id", "start_date", "end_date",
    "release_date", "archived", "frequency_code"
  )

  is.data.frame(catalogue) &&
    identical(names(catalogue), required) &&
    is.character(catalogue$title_eng) &&
    is.character(catalogue$title_fra) &&
    is.character(catalogue$id) &&
    inherits(catalogue$start_date, "Date") &&
    inherits(catalogue$end_date, "Date") &&
    inherits(catalogue$release_date, "Date") &&
    is.logical(catalogue$archived) &&
    is.integer(catalogue$frequency_code) &&
    !anyNA(catalogue$id) &&
    all(grepl("^[0-9]{2}-[0-9]{2}-[0-9]{4}-01$", catalogue$id))
}

test_that("catalogue responses are normalized", {
  payload <- data.frame(
    productId = c(10100004, 18100004),
    cubeTitleEn = c("Bank assets", "Consumer prices"),
    cubeTitleFr = c("Actifs bancaires", "Prix à la consommation"),
    cubeStartDate = c("1978-01-01", "2000-01-01"),
    cubeEndDate = c("2026-01-01", "2026-01-01"),
    releaseTime = c("2026-07-16T08:30:00Z", "2026-07-15T08:30:00Z"),
    archived = c("2", "1"),
    frequencyCode = c(12, 6)
  )

  result <- statcanR:::normalize_statcan_catalogue(payload)

  expect_identical(result$id, c("10-10-0004-01", "18-10-0004-01"))
  expect_s3_class(result$release_date, "Date")
  expect_identical(result$archived, c(FALSE, TRUE))
  expect_identical(result$frequency_code, c(12L, 6L))
  expect_true(statcanR:::is_valid_statcan_catalogue(result))
})

test_that("wrapped catalogue responses are normalized", {
  payload <- list(
    status = "SUCCESS",
    object = data.frame(
      productId = "1010000401",
      cubeTitleEn = "Bank assets",
      cubeTitleFr = "Actifs bancaires",
      releaseTime = "2026-07-16T08:30:00Z"
    )
  )

  result <- statcanR:::normalize_statcan_catalogue(payload)
  expect_identical(result$id, "10-10-0004-01")
})

test_that("catalogue title entities are decoded", {
  payload <- data.frame(
    productId = 27100363,
    cubeTitleEn = "Research &amp; development (R&amp;D)",
    cubeTitleFr = "Recherche &amp; développement (R&amp;D)",
    releaseTime = "2026-07-17T08:30:00Z"
  )

  result <- statcanR:::normalize_statcan_catalogue(payload)

  expect_identical(result$title_eng, "Research & development (R&D)")
  expect_identical(result$title_fra, "Recherche & développement (R&D)")
})

test_that("title entities in an existing catalogue cache are decoded", {
  cached <- sample_catalogue()
  cached$title_eng[[1L]] <- "Research &amp; development"
  cache_file <- tempfile(fileext = ".rds")
  on.exit(unlink(cache_file), add = TRUE)
  saveRDS(cached, cache_file)

  result <- statcanR:::read_cached_catalogue(cache_file)

  expect_identical(result$title_eng[[1L]], "Research & development")
  expect_true(statcanR:::is_valid_statcan_catalogue(result))
})

test_that("search filtering is case-insensitive", {
  result <- statcanR:::filter_statcan_catalogue(
    sample_catalogue(),
    c("INTERNATIONAL", "exports"),
    "eng"
  )

  expect_equal(nrow(result), 1L)
  expect_identical(result$id, "36-10-0434-01")
})

test_that("French titles can be searched", {
  result <- statcanR:::filter_statcan_catalogue(
    sample_catalogue(),
    "CONSOMMATION",
    "fra"
  )

  expect_equal(nrow(result), 1L)
  expect_identical(result$id, "18-10-0004-01")
  expect_identical(result$lang, "fra")
})

test_that("no matches return an empty data frame", {
  result <- statcanR:::filter_statcan_catalogue(
    sample_catalogue(),
    "not present",
    "eng"
  )

  expect_s3_class(result, "data.frame")
  expect_equal(nrow(result), 0L)
  expect_identical(
    names(result),
    c("title", "id", "description", "release_date", "lang")
  )
})

test_that("invalid catalogue schemas are rejected", {
  invalid <- sample_catalogue()
  invalid$id[1] <- "not-a-table-id"

  expect_false(statcanR:::is_valid_statcan_catalogue(invalid))
})

sample_catalogue <- function() {
  data.frame(
    title_eng = c(
      "Bank assets",
      "Consumer prices",
      "International exports",
      "Gross domestic expenditures on research and development",
      "Research and development personnel"
    ),
    title_fra = c(
      "Actifs bancaires",
      "Prix à la consommation",
      "Exportations internationales",
      "Dépenses intérieures brutes en recherche et développement",
      "Personnel affecté à la recherche et au développement"
    ),
    id = c(
      "10-10-0004-01", "18-10-0004-01", "36-10-0434-01",
      "27-10-0273-01", "27-10-0015-01"
    ),
    start_date = as.Date(c(
      "1978-01-01", "2000-01-01", "1990-01-01", "1963-01-01",
      "2015-01-01"
    )),
    end_date = as.Date(c(
      "2026-01-01", "2026-01-01", "2026-01-01", "2025-01-01",
      "2025-01-01"
    )),
    release_date = as.Date(c(
      "2026-07-16", "2026-07-15", "2026-07-14", "2025-11-27",
      "2025-08-27"
    )),
    archived = rep(FALSE, 5L),
    frequency_code = as.integer(c(12, 6, 12, 12, 12)),
    stringsAsFactors = FALSE
  )
}

#' Get an LLM's help interpreting a natural-language table search
#'
#' Sends the query and the ranked candidates from [statcan_find()] to a
#' user-configured language-model provider, which explains which
#' candidate(s) best match and asks a clarifying question when the query is
#' ambiguous. The candidate table numbers and rankings always come from
#' [statcan_find()] itself; the language model only interprets and explains
#' them, and is never allowed to propose a table number of its own.
#'
#' Two providers ship built in, selected with the `provider` argument:
#'
#' * `"openai"` (the default): the OpenAI chat-completions format, using an
#'   `Authorization: Bearer` API key. This also covers any OpenAI-compatible
#'   server -- Groq, Together, OpenRouter, Mistral, vLLM, or a local
#'   open-source model served by Ollama or LM Studio -- by pointing
#'   `endpoint` at it (a loopback `http://localhost` endpoint is accepted for
#'   local models; pass any placeholder `api_key` for servers that ignore
#'   it).
#' * `"anthropic"`: the Claude Messages format, using an `x-api-key` header.
#'
#' This is an optional feature. It requires no additional packages beyond
#' what statcanR already imports, but it does require you to configure an
#' API key and model (the endpoint defaults to the chosen provider), either
#' as arguments or through `options()` / environment variables:
#'
#' * `endpoint`: `options(statcanR.llm_endpoint = ...)` or
#'   `Sys.setenv(STATCANR_LLM_ENDPOINT = ...)`. Defaults to the provider's
#'   own endpoint when unset.
#' * `api_key`: `Sys.setenv(STATCANR_LLM_API_KEY = ...)` (or the `api_key`
#'   argument, or the provider's native variable -- `OPENAI_API_KEY` /
#'   `ANTHROPIC_API_KEY`). Because it is a secret, the key is **not** read
#'   from `options()`, which can be dumped, saved with a session, or
#'   recorded in `.Rhistory`.
#' * `model`: `options(statcanR.llm_model = ...)` or
#'   `Sys.setenv(STATCANR_LLM_MODEL = ...)`
#'
#' The endpoint must use `https://` so the key is never sent in cleartext;
#' plain `http://` is accepted only for loopback hosts (for example,
#' `http://localhost` for a local model).
#'
#' No network request is made unless `statcan_chat()` is called directly.
#'
#' @param query One non-empty character string describing the desired data.
#'   Passed to [statcan_find()].
#' @param lang Language of the table titles and of the model's reply:
#'   `"eng"` or `"fra"`.
#' @param n Maximum number of candidates to request from [statcan_find()].
#' @param refresh Logical; forwarded to [statcan_find()].
#' @param endpoint Provider endpoint URL. Defaults to
#'   `getOption("statcanR.llm_endpoint")`, then
#'   `Sys.getenv("STATCANR_LLM_ENDPOINT")`, then the chosen provider's own
#'   endpoint. Must be `https://`, except for loopback hosts (for example,
#'   `http://localhost`).
#' @param api_key API key. Defaults to `Sys.getenv("STATCANR_LLM_API_KEY")`,
#'   then the provider's native variable (`OPENAI_API_KEY` for `"openai"`,
#'   `ANTHROPIC_API_KEY` for `"anthropic"`). For safety it is not read from
#'   `options()`. It is sent as an `Authorization: Bearer` header for
#'   `"openai"` and as an `x-api-key` header for `"anthropic"`.
#' @param model Model name sent to the provider (for example, `"gpt-4o-mini"`
#'   for OpenAI or `"claude-opus-4-8"` for Anthropic). Defaults to
#'   `getOption("statcanR.llm_model")`, then
#'   `Sys.getenv("STATCANR_LLM_MODEL")`.
#' @param provider Which LLM provider to use: `"openai"` (the default,
#'   also covering OpenAI-compatible and local servers) or `"anthropic"`
#'   (Claude).
#'
#' @return A `statcan_chat_result` object: a list with `query`, `candidates`
#'   (the [statcan_find()] data frame), `explanation`, and
#'   `clarifying_question` (`NA` when the model had none).
#' @export
#'
#' @examples
#' \dontrun{
#' # OpenAI (the default provider)
#' Sys.setenv(OPENAI_API_KEY = "sk-...")
#' result <- statcan_chat(
#'   "R&D expenditures in Quebec since 2020",
#'   model = "gpt-4o-mini"
#' )
#'
#' # Anthropic (Claude)
#' Sys.setenv(ANTHROPIC_API_KEY = "sk-ant-...")
#' result <- statcan_chat(
#'   "R&D expenditures in Quebec since 2020",
#'   provider = "anthropic", model = "claude-opus-4-8"
#' )
#'
#' # A local open-source model served by Ollama (no key over loopback http)
#' result <- statcan_chat(
#'   "R&D expenditures in Quebec since 2020",
#'   endpoint = "http://localhost:11434/v1/chat/completions",
#'   api_key = "ollama", model = "llama3.1"
#' )
#'
#' # statcan_chat() returns several ranked candidates, not a single table:
#' # the model explains them but never picks or invents one for you.
#' # The candidates are already a data frame, so you never retype an id.
#' result$candidates          # the full statcan_find() data frame
#' result$candidates$id       # every candidate id, best-ranked first
#' result$candidates$id[1]    # just the top-ranked id
#' result$explanation         # the model's plain-language explanation
#'
#' # Feed the chosen id straight into statcan_data() -- nothing copied by hand.
#' table_data <- statcan_data(result$candidates$id[1], lang = "eng")
#' }
statcan_chat <- function(query, lang = c("eng", "fra"), n = 5L,
                         refresh = FALSE, endpoint = NULL, api_key = NULL,
                         model = NULL, provider = c("openai", "anthropic")) {
  lang <- match.arg(lang)
  provider <- match.arg(provider)
  config <- resolve_llm_config(endpoint, api_key, model, provider = provider)
  candidates <- statcan_find(query, lang = lang, n = n, refresh = refresh)

  if (!nrow(candidates)) {
    # No candidates: there is nothing for the model to explain, so we skip the
    # network call and return a result with no conversation state. Continuing
    # such a result is a clear error in statcan_chat_continue().
    contract <- list(
      explanation = if (lang == "eng") {
        "No matching tables were found for this query."
      } else {
        "Aucun tableau correspondant n'a \u00e9t\u00e9 trouv\u00e9 pour cette requ\u00eate."
      },
      clarifying_question = NA_character_
    )
    return(new_statcan_chat_result(query, candidates, contract, conversation = NULL))
  }

  messages <- build_llm_prompt(query, candidates, lang)
  parsed <- call_llm_chat(
    config$provider, config$endpoint, config$api_key, config$model, messages
  )
  reply <- parse_llm_reply(parsed, config$provider)
  contract <- parse_chat_contract(reply)
  messages <- c(messages, list(list(role = "assistant", content = reply)))

  new_statcan_chat_result(
    query = query,
    candidates = candidates,
    contract = contract,
    conversation = list(
      provider = provider,
      endpoint = config$endpoint,
      model = config$model,
      lang = lang,
      messages = messages
    )
  )
}


# Shared constructor for the statcan_chat_result object. `conversation` holds
# everything needed to continue the chat -- provider name, endpoint, model,
# lang, and the running message history -- but deliberately NOT the API key,
# which is a secret and must not be persisted in a saved object. It is NULL
# when there is no conversation to continue (an empty search).
new_statcan_chat_result <- function(query, candidates, contract,
                                    conversation = NULL) {
  structure(
    list(
      query = query,
      candidates = candidates,
      explanation = contract$explanation,
      clarifying_question = contract$clarifying_question,
      conversation = conversation
    ),
    class = "statcan_chat_result"
  )
}


#' Continue a `statcan_chat()` conversation
#'
#' Sends a follow-up `message` -- typically an answer to the
#' `clarifying_question` from a previous turn -- and returns an updated
#' result. The follow-up stays scoped to the **same candidate tables** that
#' [statcan_chat()] already found: it never re-runs [statcan_find()] and the
#' model still never proposes a table number of its own. To search the
#' catalogue again, start a new conversation with [statcan_chat()].
#'
#' The returned object is itself continuable, so you can chain several
#' follow-ups. The API key is re-resolved on each call (from the `api_key`
#' argument, the `STATCANR_LLM_API_KEY` environment variable, or the
#' provider's native variable) because, for safety, it is never stored in the
#' result object.
#'
#' @param result A `statcan_chat_result` returned by [statcan_chat()] (or by a
#'   previous `statcan_chat_continue()` call).
#' @param message One non-empty character string: your follow-up to the model.
#' @param api_key API key, re-resolved exactly as in [statcan_chat()]. Defaults
#'   to `Sys.getenv("STATCANR_LLM_API_KEY")`, then the provider's native
#'   variable. Not read from `options()`.
#'
#' @return An updated `statcan_chat_result` with the same `candidates`, a new
#'   `explanation` and `clarifying_question`, and an extended conversation.
#' @export
#'
#' @examples
#' \dontrun{
#' Sys.setenv(ANTHROPIC_API_KEY = "sk-ant-...")
#' chat <- statcan_chat(
#'   "R&D spending in Quebec",
#'   provider = "anthropic", model = "claude-opus-4-8"
#' )
#' chat$clarifying_question
#'
#' # Answer it and keep the same shortlist of candidates. Reassigning the same
#' # variable keeps the latest turn; you never need r1/r2/r3-style names.
#' chat <- statcan_chat_continue(chat, "annual data, since 2015")
#' chat$explanation
#'
#' # Chain another follow-up the same way:
#' chat <- statcan_chat_continue(chat, "just the total, not by industry")
#' }
statcan_chat_continue <- function(result, message, api_key = NULL) {
  if (!inherits(result, "statcan_chat_result")) {
    stop(
      "statcan_chat_continue() expects a statcan_chat_result from ",
      "statcan_chat().",
      call. = FALSE
    )
  }
  if (is.null(result$conversation)) {
    stop(
      "statcan_chat_continue(): there is no conversation to continue; the ",
      "initial statcan_chat() search returned no candidate tables.",
      call. = FALSE
    )
  }
  if (!is.character(message) || length(message) != 1L ||
        is.na(message) || !nzchar(trimws(message))) {
    stop(
      "statcan_chat_continue() requires a single non-empty message.",
      call. = FALSE
    )
  }

  conv <- result$conversation
  # Reuse the same resolution path as statcan_chat() so the stale-option
  # warning, the key fallback order, and the https-only check all behave
  # identically. Only the key is re-resolved; endpoint/model/provider come
  # from the stored conversation.
  config <- resolve_llm_config(
    conv$endpoint, api_key, conv$model, provider = conv$provider
  )

  messages <- c(conv$messages, list(list(role = "user", content = message)))
  parsed <- call_llm_chat(
    config$provider, config$endpoint, config$api_key, config$model, messages
  )
  reply <- parse_llm_reply(parsed, config$provider)
  contract <- parse_chat_contract(reply)
  messages <- c(messages, list(list(role = "assistant", content = reply)))

  new_statcan_chat_result(
    query = result$query,
    candidates = result$candidates,
    contract = contract,
    conversation = list(
      provider = conv$provider,
      endpoint = conv$endpoint,
      model = conv$model,
      lang = conv$lang,
      messages = messages
    )
  )
}


#' @export
print.statcan_chat_result <- function(x, ...) {
  cat("Query:", x$query, "\n\n")
  print(x$candidates[, c("rank", "title", "id")])
  cat("\n", x$explanation, "\n", sep = "")
  if (!is.na(x$clarifying_question)) {
    cat("\nClarifying question:", x$clarifying_question, "\n")
  }
  invisible(x)
}


# The built-in provider registry. Each provider bundles the few things that
# actually differ between vendors: the default endpoint, the native API-key
# environment variable, how the API key is attached (auth scheme), how the
# request body is shaped, and how the reply is read back out. Everything else
# -- the statcan_find() ranking, the prompt text, the reply contract, and the
# https-only security check -- is provider-neutral.
llm_provider <- function(provider = c("openai", "anthropic")) {
  provider <- match.arg(provider)
  switch(
    provider,
    openai = list(
      name = "openai",
      endpoint = "https://api.openai.com/v1/chat/completions",
      key_env = "OPENAI_API_KEY",
      build_headers = function(api_key) {
        httr::add_headers(Authorization = paste("Bearer", api_key))
      },
      build_body = function(model, messages) {
        list(model = model, messages = messages)
      },
      parse_content = function(parsed) {
        parsed$choices[[1L]]$message$content
      }
    ),
    anthropic = list(
      name = "anthropic",
      endpoint = "https://api.anthropic.com/v1/messages",
      key_env = "ANTHROPIC_API_KEY",
      build_headers = function(api_key) {
        httr::add_headers(
          `x-api-key` = api_key,
          `anthropic-version` = "2023-06-01"
        )
      },
      build_body = function(model, messages) {
        # Claude takes the system prompt as a top-level field, not a message,
        # and requires max_tokens. Lift the neutral "system" message out and
        # keep the rest (the user turn) as messages.
        system_text <- NULL
        chat <- list()
        for (message in messages) {
          if (identical(message$role, "system")) {
            system_text <- message$content
          } else {
            chat[[length(chat) + 1L]] <- message
          }
        }
        body <- list(model = model, max_tokens = 4096L, messages = chat)
        if (!is.null(system_text)) {
          body$system <- system_text
        }
        body
      },
      parse_content = function(parsed) {
        parsed$content[[1L]]$text
      }
    )
  )
}


resolve_llm_config <- function(endpoint, api_key, model,
                               provider = c("openai", "anthropic")) {
  provider <- match.arg(provider)
  prov <- llm_provider(provider)

  endpoint <- first_nonempty(
    endpoint,
    getOption("statcanR.llm_endpoint"),
    Sys.getenv("STATCANR_LLM_ENDPOINT")
  )
  if (is.null(endpoint)) {
    endpoint <- prov$endpoint
  }
  # The API key is a secret, so it is deliberately not read from options():
  # options can be dumped with options(), captured in a saved session, or land
  # in .Rhistory when set inline. It comes only from the api_key argument, the
  # STATCANR_LLM_API_KEY environment variable, or the provider's native key
  # variable (OPENAI_API_KEY / ANTHROPIC_API_KEY). Earlier versions did read the
  # option, so warn if a stale one is set. Only its presence is checked with
  # nzchar(); the value itself is never read, so the key is not re-exposed.
  if (nzchar(getOption("statcanR.llm_api_key", ""))) {
    warning(
      "The statcanR.llm_api_key option is ignored for security. Set the API ",
      "key via Sys.setenv(STATCANR_LLM_API_KEY = ) or the api_key argument ",
      "instead.",
      call. = FALSE
    )
  }
  api_key <- first_nonempty(
    api_key,
    Sys.getenv("STATCANR_LLM_API_KEY"),
    Sys.getenv(prov$key_env)
  )
  model <- first_nonempty(
    model,
    getOption("statcanR.llm_model"),
    Sys.getenv("STATCANR_LLM_MODEL")
  )

  if (is.null(endpoint) || is.null(api_key) || is.null(model)) {
    stop(
      "statcan_chat() requires an LLM endpoint, API key, and model. The ",
      "endpoint defaults to the chosen provider, or set it via the endpoint ",
      "argument, options(statcanR.llm_endpoint = ), or STATCANR_LLM_ENDPOINT. ",
      "Set the model via the model argument, options(statcanR.llm_model = ), ",
      "or STATCANR_LLM_MODEL. Set the API key via the api_key argument, ",
      "Sys.setenv(STATCANR_LLM_API_KEY = ), or the provider's native ",
      "variable (OPENAI_API_KEY / ANTHROPIC_API_KEY); for safety it is not ",
      "read from options().",
      call. = FALSE
    )
  }

  validate_llm_endpoint(endpoint)

  list(provider = prov, endpoint = endpoint, api_key = api_key, model = model)
}


# Refuse to send the API key over an unencrypted connection. https is always
# allowed; plain http is allowed only for loopback hosts, so that local models
# (for example, an Ollama server on http://localhost) keep working while a
# mistyped public http:// endpoint cannot leak the key in cleartext.
validate_llm_endpoint <- function(endpoint) {
  parsed <- httr::parse_url(endpoint)
  scheme <- tolower(if (is.null(parsed$scheme)) "" else parsed$scheme)
  host <- tolower(if (is.null(parsed$hostname)) "" else parsed$hostname)

  if (scheme == "https") {
    return(invisible(endpoint))
  }

  is_loopback <- host %in% c("localhost", "127.0.0.1", "::1", "[::1]") ||
    startsWith(host, "127.")
  if (scheme == "http" && is_loopback) {
    return(invisible(endpoint))
  }

  if (scheme == "http") {
    stop(
      "statcan_chat() will not send the API key to an unencrypted http:// ",
      "endpoint (", host, "). Use an https:// endpoint, or a loopback host ",
      "such as http://localhost for a local model.",
      call. = FALSE
    )
  }

  stop(
    "statcan_chat() requires an https:// endpoint (or http:// on a loopback ",
    "host for a local model). Received: ", endpoint, ".",
    call. = FALSE
  )
}


first_nonempty <- function(...) {
  for (value in list(...)) {
    if (is.null(value) || length(value) != 1L || is.na(value)) {
      next
    }
    if (nzchar(value)) {
      return(value)
    }
  }
  NULL
}


build_llm_prompt <- function(query, candidates, lang) {
  candidate_lines <- paste0(
    "- ", candidates$id, ": \"", candidates$title, "\"",
    " (", candidates$start_date, " to ", candidates$end_date, ")"
  )
  candidate_block <- paste(candidate_lines, collapse = "\n")

  if (lang == "eng") {
    system_content <- paste(
      "You help a user pick the right Statistics Canada data table.",
      "You are given a user query and a ranked list of candidate tables",
      "with their official table numbers. You must only ever refer to",
      "the table numbers given to you; never invent or guess a table",
      "number. Reply in English using exactly this format, with no other",
      "text before or after it:\n",
      "EXPLANATION:\n<one or two short paragraphs>\n",
      "CLARIFYING_QUESTION:\n<a question, or the literal word NONE>"
    )
  } else {
    system_content <- paste(
      "Vous aidez un utilisateur \u00e0 choisir le bon tableau de donn\u00e9es",
      "de Statistique Canada. On vous donne une requ\u00eate et une liste",
      "class\u00e9e de tableaux candidats avec leurs num\u00e9ros officiels. Vous ne",
      "devez jamais inventer ou deviner un num\u00e9ro de tableau. R\u00e9pondez en",
      "fran\u00e7ais en utilisant exactement ce format, sans aucun autre",
      "texte avant ou apr\u00e8s :\n",
      "EXPLANATION:\n<un ou deux courts paragraphes>\n",
      "CLARIFYING_QUESTION:\n<une question, ou le mot NONE>"
    )
  }

  user_content <- paste0(
    "Query: ", query, "\n\nCandidate tables:\n", candidate_block
  )

  list(
    list(role = "system", content = system_content),
    list(role = "user", content = user_content)
  )
}


call_llm_chat <- function(provider, endpoint, api_key, model, messages) {
  response <- tryCatch(
    httr::POST(
      endpoint,
      provider$build_headers(api_key),
      body = provider$build_body(model, messages),
      encode = "json",
      httr::timeout(60),
      httr::user_agent(statcan_user_agent())
    ),
    error = function(error) {
      stop(
        "statcan_chat(): unable to contact the LLM endpoint. ",
        conditionMessage(error),
        call. = FALSE
      )
    }
  )
  stop_for_llm_status(response)

  tryCatch(
    jsonlite::fromJSON(
      httr::content(response, as = "text", encoding = "UTF-8"),
      simplifyVector = FALSE
    ),
    error = function(error) {
      stop(
        "statcan_chat(): the LLM endpoint returned invalid JSON. ",
        conditionMessage(error),
        call. = FALSE
      )
    }
  )
}


stop_for_llm_status <- function(response) {
  if (!httr::http_error(response)) {
    return(invisible(response))
  }

  status <- httr::status_code(response)
  detail <- tryCatch(
    httr::content(response, as = "text", encoding = "UTF-8"),
    error = function(error) ""
  )
  detail <- trimws(gsub("[\r\n]+", " ", detail))
  if (nchar(detail) > 200L) {
    detail <- paste0(substr(detail, 1L, 200L), "...")
  }
  suffix <- if (nzchar(detail)) paste0(" ", detail) else ""

  stop(
    "statcan_chat(): the LLM request failed with HTTP ", status, ".",
    suffix,
    call. = FALSE
  )
}


parse_llm_reply <- function(parsed, provider = llm_provider("openai")) {
  content <- tryCatch(
    provider$parse_content(parsed),
    error = function(error) NULL
  )
  if (is.null(content) || length(content) != 1L || !nzchar(content)) {
    stop(
      "statcan_chat(): the LLM endpoint's response did not include a ",
      "chat completion message.",
      call. = FALSE
    )
  }
  content
}


parse_chat_contract <- function(text) {
  match <- regmatches(
    text,
    regexec(
      "(?s)EXPLANATION:\\s*(.*?)\\s*CLARIFYING_QUESTION:\\s*(.*)",
      text,
      perl = TRUE
    )
  )[[1L]]

  if (length(match) != 3L) {
    return(list(
      explanation = trimws(text),
      clarifying_question = NA_character_
    ))
  }

  clarifying_question <- trimws(match[[3L]])
  list(
    explanation = trimws(match[[2L]]),
    clarifying_question = if (toupper(clarifying_question) == "NONE") {
      NA_character_
    } else {
      clarifying_question
    }
  )
}

sample_candidates <- function() {
  data.frame(
    rank = 1L,
    title = "Gross domestic expenditures on research and development",
    id = "27-10-0273-01",
    score = 90,
    match_reason = "3/3 topic terms matched",
    start_date = as.Date("1963-01-01"),
    end_date = as.Date("2025-01-01"),
    release_date = as.Date("2025-11-27"),
    geography = NA_character_,
    geography_match = NA,
    lang = "eng",
    stringsAsFactors = FALSE
  )
}


test_that("resolve_llm_config resolves endpoint and model via arg, option, env", {
  withr::local_options(
    statcanR.llm_endpoint = "https://option-endpoint",
    statcanR.llm_model = "option-model"
  )
  withr::local_envvar(
    STATCANR_LLM_ENDPOINT = "https://env-endpoint",
    STATCANR_LLM_API_KEY = "env-key",
    STATCANR_LLM_MODEL = "env-model"
  )

  # Argument beats option beats environment variable.
  config <- statcanR:::resolve_llm_config("https://arg-endpoint", NULL, NULL)
  expect_identical(config$endpoint, "https://arg-endpoint")
  expect_identical(config$api_key, "env-key")
  expect_identical(config$model, "option-model")

  withr::local_options(
    statcanR.llm_endpoint = NULL,
    statcanR.llm_model = NULL
  )
  config <- statcanR:::resolve_llm_config(NULL, NULL, NULL)
  expect_identical(config$endpoint, "https://env-endpoint")
  expect_identical(config$api_key, "env-key")
  expect_identical(config$model, "env-model")
})


test_that("a stale statcanR.llm_api_key option is ignored with a warning", {
  withr::local_options(
    statcanR.llm_endpoint = "https://endpoint",
    statcanR.llm_api_key = "option-key",
    statcanR.llm_model = "model"
  )
  withr::local_envvar(
    STATCANR_LLM_API_KEY = NA,
    OPENAI_API_KEY = NA,
    ANTHROPIC_API_KEY = NA
  )

  # The option is ignored, but a warning guides migration; the explicit
  # argument still supplies the key.
  expect_warning(
    config <- statcanR:::resolve_llm_config(NULL, "arg-key", NULL),
    "ignored for security"
  )
  expect_identical(config$api_key, "arg-key")

  # With no argument and no env var, the ignored option leaves the key missing.
  expect_error(
    suppressWarnings(statcanR:::resolve_llm_config(NULL, NULL, NULL)),
    "requires an LLM endpoint"
  )
})


test_that("validate_llm_endpoint enforces https except on loopback hosts", {
  expect_silent(
    statcanR:::validate_llm_endpoint("https://api.openai.com/v1/chat")
  )
  expect_silent(
    statcanR:::validate_llm_endpoint("http://localhost:11434/v1/chat")
  )
  expect_silent(
    statcanR:::validate_llm_endpoint("http://127.0.0.1:11434/v1/chat")
  )

  expect_error(
    statcanR:::validate_llm_endpoint("http://api.example.com/v1/chat"),
    "unencrypted"
  )
  expect_error(
    statcanR:::validate_llm_endpoint("ftp://example.com/x"),
    "requires an https"
  )
})


test_that("resolve_llm_config errors clearly when nothing is configured", {
  old_options <- options(
    statcanR.llm_endpoint = NULL,
    statcanR.llm_api_key = NULL,
    statcanR.llm_model = NULL
  )
  old_env <- Sys.getenv(
    c(
      "STATCANR_LLM_ENDPOINT", "STATCANR_LLM_API_KEY", "STATCANR_LLM_MODEL",
      "OPENAI_API_KEY", "ANTHROPIC_API_KEY"
    ),
    unset = NA
  )
  Sys.setenv(
    STATCANR_LLM_ENDPOINT = "", STATCANR_LLM_API_KEY = "",
    STATCANR_LLM_MODEL = "", OPENAI_API_KEY = "", ANTHROPIC_API_KEY = ""
  )
  on.exit({
    options(old_options)
    for (name in names(old_env)) {
      if (is.na(old_env[[name]])) Sys.unsetenv(name) else {
        do.call(Sys.setenv, stats::setNames(list(old_env[[name]]), name))
      }
    }
  })

  expect_error(
    statcanR:::resolve_llm_config(NULL, NULL, NULL),
    "requires an LLM endpoint"
  )
})


test_that("resolve_llm_config defaults the endpoint and key var per provider", {
  withr::local_options(
    statcanR.llm_endpoint = NULL,
    statcanR.llm_model = NULL
  )
  withr::local_envvar(
    STATCANR_LLM_ENDPOINT = NA,
    STATCANR_LLM_API_KEY = NA,
    STATCANR_LLM_MODEL = NA,
    OPENAI_API_KEY = "openai-key",
    ANTHROPIC_API_KEY = "anthropic-key"
  )

  # openai: default endpoint, key from OPENAI_API_KEY
  openai <- statcanR:::resolve_llm_config(NULL, NULL, "gpt-4o-mini")
  expect_identical(openai$provider$name, "openai")
  expect_identical(openai$endpoint, "https://api.openai.com/v1/chat/completions")
  expect_identical(openai$api_key, "openai-key")

  # anthropic: default endpoint, key from ANTHROPIC_API_KEY
  anthropic <- statcanR:::resolve_llm_config(
    NULL, NULL, "claude-opus-4-8", provider = "anthropic"
  )
  expect_identical(anthropic$provider$name, "anthropic")
  expect_identical(anthropic$endpoint, "https://api.anthropic.com/v1/messages")
  expect_identical(anthropic$api_key, "anthropic-key")
})


test_that("the anthropic provider shapes the body and parses the reply", {
  prov <- statcanR:::llm_provider("anthropic")

  messages <- list(
    list(role = "system", content = "system text"),
    list(role = "user", content = "user text")
  )
  body <- prov$build_body("claude-opus-4-8", messages)

  # system is lifted to a top-level field; only the user turn stays in messages
  expect_identical(body$system, "system text")
  expect_length(body$messages, 1L)
  expect_identical(body$messages[[1L]]$role, "user")
  # Claude requires max_tokens
  expect_true(is.numeric(body$max_tokens))

  # reply content lives at content[[1]]$text, not choices[[1]]$message$content
  parsed <- list(content = list(list(type = "text", text = "hello from claude")))
  expect_identical(prov$parse_content(parsed), "hello from claude")
  expect_identical(
    statcanR:::parse_llm_reply(parsed, prov), "hello from claude"
  )
})


test_that("build_llm_prompt embeds the query and candidate table numbers", {
  messages <- statcanR:::build_llm_prompt(
    "R&D expenditures in Quebec since 2020", sample_candidates(), "eng"
  )
  expect_length(messages, 2L)
  expect_identical(messages[[1L]]$role, "system")
  expect_identical(messages[[2L]]$role, "user")
  expect_match(messages[[2L]]$content, "27-10-0273-01", fixed = TRUE)
  expect_match(
    messages[[2L]]$content, "R&D expenditures in Quebec since 2020",
    fixed = TRUE
  )
})


test_that("parse_llm_reply extracts the chat completion message", {
  parsed <- list(choices = list(list(message = list(content = "hello"))))
  expect_identical(statcanR:::parse_llm_reply(parsed), "hello")

  expect_error(
    statcanR:::parse_llm_reply(list(choices = list())),
    "did not include a chat completion message"
  )
})


test_that("parse_chat_contract splits a well-formed reply", {
  text <- paste(
    "EXPLANATION:",
    "Table 27-10-0273-01 covers Quebec since 2020.",
    "CLARIFYING_QUESTION:",
    "NONE",
    sep = "\n"
  )
  result <- statcanR:::parse_chat_contract(text)
  expect_match(result$explanation, "27-10-0273-01", fixed = TRUE)
  expect_true(is.na(result$clarifying_question))
})


test_that("parse_chat_contract keeps a real clarifying question", {
  text <- paste(
    "EXPLANATION:",
    "Two tables could work.",
    "CLARIFYING_QUESTION:",
    "Do you need monthly or annual data?",
    sep = "\n"
  )
  result <- statcanR:::parse_chat_contract(text)
  expect_identical(
    result$clarifying_question, "Do you need monthly or annual data?"
  )
})


test_that("parse_chat_contract falls back gracefully on a malformed reply", {
  result <- statcanR:::parse_chat_contract("just some free text")
  expect_identical(result$explanation, "just some free text")
  expect_true(is.na(result$clarifying_question))
})


test_that("statcan_chat skips the LLM call when statcan_find has no candidates", {
  local_mocked_bindings(
    statcan_find = function(...) statcanR:::empty_statcan_find_result(),
    call_llm_chat = function(...) stop("the LLM should not be called"),
    .package = "statcanR"
  )

  result <- statcan_chat(
    "an unmatched query", endpoint = "https://x", api_key = "k",
    model = "m"
  )
  expect_s3_class(result, "statcan_chat_result")
  expect_identical(nrow(result$candidates), 0L)
  expect_match(result$explanation, "No matching tables")
  expect_true(is.na(result$clarifying_question))
})


test_that("statcan_chat assembles a result from the mocked LLM reply", {
  local_mocked_bindings(
    statcan_find = function(...) sample_candidates(),
    call_llm_chat = function(...) {
      list(choices = list(list(message = list(content = paste(
        "EXPLANATION:",
        "Table 27-10-0273-01 is the best match.",
        "CLARIFYING_QUESTION:",
        "NONE",
        sep = "\n"
      )))))
    },
    .package = "statcanR"
  )

  result <- statcan_chat(
    "R&D expenditures in Quebec since 2020",
    endpoint = "https://x", api_key = "k", model = "m"
  )
  expect_s3_class(result, "statcan_chat_result")
  expect_identical(result$candidates, sample_candidates())
  expect_match(result$explanation, "27-10-0273-01", fixed = TRUE)
  expect_true(is.na(result$clarifying_question))
})


# The mocked endpoint echoes how many messages it was sent, so tests can assert
# that the conversation grows by the expected amount on each turn.
echo_message_count <- function(provider, endpoint, api_key, model, messages) {
  list(choices = list(list(message = list(content = paste0(
    "EXPLANATION:\n", length(messages), " messages\nCLARIFYING_QUESTION:\nNONE"
  )))))
}


test_that("statcan_chat stores conversation state for a successful turn", {
  local_mocked_bindings(
    statcan_find = function(...) sample_candidates(),
    call_llm_chat = echo_message_count,
    .package = "statcanR"
  )

  r1 <- statcan_chat("q", endpoint = "https://x", api_key = "k", model = "m")
  expect_false(is.null(r1$conversation))
  expect_identical(r1$conversation$provider, "openai")
  # system + user sent on the first turn
  expect_match(r1$explanation, "2 messages")
  # stored history also includes the assistant reply -> 3 turns
  expect_length(r1$conversation$messages, 3L)
})


test_that("statcan_chat_continue stays scoped to candidates and chains", {
  local_mocked_bindings(
    statcan_find = function(...) sample_candidates(),
    call_llm_chat = echo_message_count,
    .package = "statcanR"
  )

  r1 <- statcan_chat("q", endpoint = "https://x", api_key = "k", model = "m")

  r2 <- statcan_chat_continue(r1, "annual data", api_key = "k")
  # continue appends the user turn to the stored 3 -> 4 sent
  expect_match(r2$explanation, "4 messages")
  expect_identical(r2$candidates, r1$candidates) # shortlist unchanged

  # chainable: a second follow-up works on the result of the first
  r3 <- statcan_chat_continue(r2, "since 2015", api_key = "k")
  expect_match(r3$explanation, "6 messages")
  expect_length(r3$conversation$messages, 7L)

  expect_error(
    statcan_chat_continue(r1, "", api_key = "k"),
    "non-empty message"
  )
})


test_that("statcan_chat_continue errors when there is no conversation", {
  local_mocked_bindings(
    statcan_find = function(...) statcanR:::empty_statcan_find_result(),
    call_llm_chat = function(...) stop("the LLM should not be called"),
    .package = "statcanR"
  )

  r <- statcan_chat(
    "no matches", endpoint = "https://x", api_key = "k", model = "m"
  )
  expect_true(is.null(r$conversation))
  expect_error(
    statcan_chat_continue(r, "more detail", api_key = "k"),
    "no conversation to continue"
  )
})


test_that("statcan_chat_continue rejects a non-result first argument", {
  expect_error(statcan_chat_continue(list(), "hi"), "statcan_chat_result")
})

# Redirect the package cache directory into a temporary location so the test
# suite never reads from or writes to the real user cache (catalogue, catalogue
# tokens, and cube metadata all live there).
cache_home <- tempfile("statcanR-cache-")
dir.create(cache_home, recursive = TRUE, showWarnings = FALSE)
old_cache <- Sys.getenv("R_USER_CACHE_DIR", unset = NA)
Sys.setenv(R_USER_CACHE_DIR = cache_home)

withr::defer(
  {
    if (is.na(old_cache)) {
      Sys.unsetenv("R_USER_CACHE_DIR")
    } else {
      Sys.setenv(R_USER_CACHE_DIR = old_cache)
    }
    unlink(cache_home, recursive = TRUE, force = TRUE)
  },
  teardown_env()
)

#' Find Statistics Canada tables using a natural-language query
#'
#' Interprets a short description of the data needed and returns ranked
#' Statistics Canada table candidates. The query can contain a topic, Canadian
#' geography, and reference year or range. For example,
#' `"R&D expenditures in Quebec since 2020"` is interpreted as a research and
#' development expenditure topic, a Quebec geography constraint, and coverage
#' beginning in 2020.
#'
#' This function finds tables; it does not download or filter their
#' observations. Use the returned `id` with [statcan_data()], then apply any
#' required geography and date filters to the downloaded data. Rankings are a
#' discovery aid, so review the candidate title before downloading a large
#' table.
#'
#' The catalogue is cached for 24 hours. When a geography is present in the
#' query, metadata for a small set of leading candidates is retrieved through
#' WDS to confirm that the geography is a table member. That metadata is cached
#' for seven days. If metadata is temporarily unavailable, candidates can
#' still be returned with `geography_match = NA`.
#'
#' @param query One non-empty character string describing the desired data.
#' @param lang Language of the table titles: `"eng"` or `"fra"`.
#' @param n Maximum number of candidates to return, from 1 to 20.
#' @param refresh Logical; if `TRUE`, request a fresh catalogue and fresh
#'   candidate metadata from WDS.
#'
#' @return A data frame of ranked candidates. `id` contains the table number,
#'   `score` is the relevance score, and `match_reason` explains the ranking.
#'   Coverage dates describe the table as a whole. `geography_match` is `TRUE`
#'   when WDS metadata confirms every geography in the query, `FALSE` when it
#'   does not, and `NA` when no geography was requested or validation was not
#'   possible.
#' @export
#'
#' @examples
#' \dontrun{
#' matches <- statcan_find(
#'   "R&D expenditures in Quebec since 2020",
#'   lang = "eng"
#' )
#' matches[, c("title", "id", "match_reason")]
#' }
statcan_find <- function(query, lang = c("eng", "fra"), n = 5L,
                         refresh = FALSE) {
  lang <- match.arg(lang)
  if (!is.character(query) || length(query) != 1L || is.na(query) ||
      !nzchar(trimws(query))) {
    stop("`query` must be one non-empty character string.", call. = FALSE)
  }
  if (!is.numeric(n) || length(n) != 1L || is.na(n) || n %% 1 != 0 ||
      n < 1 || n > 20) {
    stop("`n` must be one whole number from 1 to 20.", call. = FALSE)
  }
  if (!is.logical(refresh) || length(refresh) != 1L || is.na(refresh)) {
    stop("`refresh` must be TRUE or FALSE.", call. = FALSE)
  }
  n <- as.integer(n)

  parsed <- parse_statcan_query(query, lang)
  if (!length(parsed$topic_terms)) {
    stop(
      "The query must include a subject, not only a geography or date.",
      call. = FALSE
    )
  }

  candidates <- rank_statcan_catalogue(
    statcan_catalogue(refresh = refresh),
    parsed,
    lang
  )
  if (!nrow(candidates)) {
    return(empty_statcan_find_result())
  }

  geography_label <- parsed_geography_label(parsed, lang)
  candidates$geography <- rep(geography_label, nrow(candidates))
  candidates$geography_match <- rep(NA, nrow(candidates))

  if (nrow(parsed$geographies)) {
    pool_size <- min(nrow(candidates), max(20L, min(60L, n * 10L)))
    candidates <- candidates[seq_len(pool_size), , drop = FALSE]
    metadata <- statcan_metadata_for_tables(
      candidates$id,
      refresh = refresh
    )
    product_ids <- vapply(
      candidates$id,
      normalize_product_id,
      character(1)
    )
    candidates$geography_match <- vapply(
      product_ids,
      function(product_id) {
        candidate_metadata <- metadata[[product_id]]
        if (is.null(candidate_metadata)) {
          return(NA)
        }
        metadata_matches_geographies(
          candidate_metadata,
          parsed$geographies
        )
      },
      logical(1)
    )

    candidates$score[candidates$geography_match %in% TRUE] <- pmin(
      100,
      candidates$score[candidates$geography_match %in% TRUE] + 10
    )
    candidates <- candidates[
      is.na(candidates$geography_match) |
        candidates$geography_match,
      ,
      drop = FALSE
    ]
    if (!nrow(candidates)) {
      return(empty_statcan_find_result())
    }
    candidates <- order_statcan_candidates(candidates)
  }

  candidates <- candidates[seq_len(min(n, nrow(candidates))), , drop = FALSE]
  candidates$match_reason <- vapply(
    seq_len(nrow(candidates)),
    function(index) build_statcan_match_reason(
      candidates[index, , drop = FALSE],
      parsed,
      lang
    ),
    character(1)
  )
  candidates$rank <- seq_len(nrow(candidates))
  candidates$lang <- rep(lang, nrow(candidates))

  result <- candidates[
    c(
      "rank", "title", "id", "score", "match_reason", "start_date",
      "end_date", "release_date", "geography", "geography_match", "lang"
    )
  ]
  rownames(result) <- NULL
  result
}


empty_statcan_find_result <- function() {
  data.frame(
    rank = integer(),
    title = character(),
    id = character(),
    score = numeric(),
    match_reason = character(),
    start_date = as.Date(character()),
    end_date = as.Date(character()),
    release_date = as.Date(character()),
    geography = character(),
    geography_match = logical(),
    lang = character(),
    stringsAsFactors = FALSE
  )
}


parse_statcan_query <- function(query, lang) {
  normalized <- normalize_statcan_search_text(query)
  geography_result <- extract_statcan_geographies(normalized)
  dates <- extract_statcan_query_dates(normalized, lang)
  topic_text <- geography_result$remaining_text
  topic_text <- gsub("\\b(?:19|20)[0-9]{2}\\b", " ", topic_text, perl = TRUE)

  list(
    topic_terms = canonical_statcan_tokens(topic_text),
    geographies = geography_result$geographies,
    start_date = dates$start_date,
    end_date = dates$end_date,
    date_label = dates$label
  )
}


normalize_statcan_search_text <- function(text) {
  text <- tolower(text)
  text <- gsub("&amp;", "&", text, fixed = TRUE)
  text <- gsub("&#0*39;|&apos;", "'", text, perl = TRUE)
  text <- gsub(
    "\\br\\s*(?:[-&+/]|and|et)\\s*d\\b",
    " research and development ",
    text,
    perl = TRUE
  )
  # `iconv(..., "ASCII//TRANSLIT")` is platform-dependent. In particular,
  # macOS can transliterate French accents with apostrophes (for example,
  # "Quebec" can become "qu'ebec"). Normalize common Latin characters first
  # so query parsing gives the same result on all CRAN platforms.
  replacements <- c(
    "[\u00e0\u00e1\u00e2\u00e3\u00e4\u00e5]" = "a",
    "[\u00e7\u0107\u010d]" = "c", "[\u010f\u00f0]" = "d",
    "[\u00e8\u00e9\u00ea\u00eb]" = "e",
    "[\u00ec\u00ed\u00ee\u00ef]" = "i", "[\u00f1\u0148]" = "n",
    "[\u00f2\u00f3\u00f4\u00f5\u00f6\u00f8]" = "o",
    "[\u0159]" = "r", "[\u0161]" = "s", "[\u0165]" = "t",
    "[\u00f9\u00fa\u00fb\u00fc]" = "u", "[\u00fd\u00ff]" = "y",
    "[\u017e]" = "z", "\u00e6" = "ae", "\u0153" = "oe"
  )
  for (pattern in names(replacements)) {
    text <- gsub(pattern, replacements[[pattern]], text, perl = TRUE)
  }
  text <- iconv(text, to = "ASCII//TRANSLIT", sub = "")
  text <- gsub("[^a-z0-9]+", " ", text)
  trimws(gsub("[[:space:]]+", " ", text))
}


canonical_statcan_tokens <- function(text) {
  normalized <- normalize_statcan_search_text(text)
  if (!nzchar(normalized)) {
    return(character())
  }
  normalized <- expand_statcan_acronyms(normalized)
  tokens <- unlist(strsplit(normalized, " ", fixed = TRUE), use.names = FALSE)
  dictionary <- c(
    research = "research",
    recherche = "research",
    recherches = "research",
    development = "development",
    developpement = "development",
    expenditure = "expenditure",
    expenditures = "expenditure",
    expense = "expenditure",
    expenses = "expenditure",
    spending = "expenditure",
    depense = "expenditure",
    depenses = "expenditure"
  )
  mapped <- unname(dictionary[tokens])
  tokens[!is.na(mapped)] <- mapped[!is.na(mapped)]

  stopwords <- c(
    "a", "about", "after", "all", "an", "and", "as", "at", "before",
    "between", "by", "data", "for", "from", "get", "give", "i", "in",
    "into", "me", "of", "on", "please", "province", "show", "since",
    "starting", "table", "tables", "the", "through", "to", "until",
    "with", "year", "years", "au", "aux", "avec", "avant", "de", "des",
    "depuis", "du", "en", "entre", "et", "jusqu", "la", "le", "les",
    "par", "pour", "sur", "un", "une", "r", "d"
  )
  unique(tokens[
    nzchar(tokens) &
      !tokens %in% stopwords &
      !grepl("^(?:19|20)[0-9]{2}$", tokens)
  ])
}


expand_statcan_acronyms <- function(text) {
  acronyms <- c(
    gdp = "gross domestic product",
    pib = "produit interieur brut",
    cpi = "consumer price index",
    ipc = "indice des prix a la consommation",
    ippi = "industrial product price index",
    ppi = "industrial product price index"
  )
  for (acronym in names(acronyms)) {
    text <- gsub(
      paste0("\\b", acronym, "\\b"),
      acronyms[[acronym]],
      text,
      perl = TRUE
    )
  }
  trimws(gsub("[[:space:]]+", " ", text))
}


statcan_geography_dictionary <- function() {
  list(
    list(eng = "Canada", fra = "Canada", aliases = c("canada")),
    list(
      eng = "Newfoundland and Labrador",
      fra = "Terre-Neuve-et-Labrador",
      aliases = c("newfoundland and labrador", "terre neuve et labrador")
    ),
    list(
      eng = "Prince Edward Island",
      fra = "\u00cele-du-Prince-\u00c9douard",
      aliases = c("prince edward island", "ile du prince edouard")
    ),
    list(
      eng = "Nova Scotia", fra = "Nouvelle-\u00c9cosse",
      aliases = c("nova scotia", "nouvelle ecosse")
    ),
    list(
      eng = "New Brunswick", fra = "Nouveau-Brunswick",
      aliases = c("new brunswick", "nouveau brunswick")
    ),
    list(
      eng = "Quebec", fra = "Qu\u00e9bec",
      aliases = c("province of quebec", "province de quebec", "quebec", "qc")
    ),
    list(eng = "Ontario", fra = "Ontario", aliases = c("ontario")),
    list(eng = "Manitoba", fra = "Manitoba", aliases = c("manitoba")),
    list(
      eng = "Saskatchewan", fra = "Saskatchewan",
      aliases = c("saskatchewan")
    ),
    list(eng = "Alberta", fra = "Alberta", aliases = c("alberta")),
    list(
      eng = "British Columbia", fra = "Colombie-Britannique",
      aliases = c("british columbia", "colombie britannique", "b c")
    ),
    list(eng = "Yukon", fra = "Yukon", aliases = c("yukon")),
    list(
      eng = "Northwest Territories", fra = "Territoires du Nord-Ouest",
      aliases = c("northwest territories", "territoires du nord ouest")
    ),
    list(eng = "Nunavut", fra = "Nunavut", aliases = c("nunavut")),
    list(eng = "St. John's", fra = "St. John's", aliases = c("st johns")),
    list(eng = "Halifax", fra = "Halifax", aliases = c("halifax")),
    list(eng = "Moncton", fra = "Moncton", aliases = c("moncton")),
    list(eng = "Saint John", fra = "Saint John", aliases = c("saint john")),
    list(eng = "Saguenay", fra = "Saguenay", aliases = c("saguenay")),
    list(
      eng = "Quebec City", fra = "Qu\u00e9bec",
      aliases = c("quebec city")
    ),
    list(eng = "Sherbrooke", fra = "Sherbrooke", aliases = c("sherbrooke")),
    list(
      eng = "Trois-Rivi\u00e8res", fra = "Trois-Rivi\u00e8res",
      aliases = c("trois rivieres")
    ),
    list(eng = "Montreal", fra = "Montr\u00e9al", aliases = c("montreal")),
    list(
      eng = "Ottawa - Gatineau", fra = "Ottawa - Gatineau",
      aliases = c("ottawa gatineau", "ottawa", "gatineau")
    ),
    list(eng = "Kingston", fra = "Kingston", aliases = c("kingston")),
    list(eng = "Belleville", fra = "Belleville", aliases = c("belleville")),
    list(
      eng = "Peterborough", fra = "Peterborough",
      aliases = c("peterborough")
    ),
    list(eng = "Oshawa", fra = "Oshawa", aliases = c("oshawa")),
    list(eng = "Toronto", fra = "Toronto", aliases = c("toronto")),
    list(eng = "Hamilton", fra = "Hamilton", aliases = c("hamilton")),
    list(
      eng = "St. Catharines - Niagara",
      fra = "St. Catharines - Niagara",
      aliases = c("st catharines niagara", "st catharines")
    ),
    list(
      eng = "Kitchener - Cambridge - Waterloo",
      fra = "Kitchener - Cambridge - Waterloo",
      aliases = c(
        "kitchener cambridge waterloo", "kitchener", "cambridge", "waterloo"
      )
    ),
    list(eng = "Brantford", fra = "Brantford", aliases = c("brantford")),
    list(eng = "Guelph", fra = "Guelph", aliases = c("guelph")),
    list(eng = "London", fra = "London", aliases = c("london")),
    list(eng = "Windsor", fra = "Windsor", aliases = c("windsor")),
    list(eng = "Barrie", fra = "Barrie", aliases = c("barrie")),
    list(
      eng = "Greater Sudbury", fra = "Grand Sudbury",
      aliases = c("greater sudbury", "grand sudbury", "sudbury")
    ),
    list(
      eng = "Thunder Bay", fra = "Thunder Bay",
      aliases = c("thunder bay")
    ),
    list(eng = "Winnipeg", fra = "Winnipeg", aliases = c("winnipeg")),
    list(eng = "Regina", fra = "Regina", aliases = c("regina")),
    list(eng = "Saskatoon", fra = "Saskatoon", aliases = c("saskatoon")),
    list(eng = "Calgary", fra = "Calgary", aliases = c("calgary")),
    list(eng = "Edmonton", fra = "Edmonton", aliases = c("edmonton")),
    list(eng = "Lethbridge", fra = "Lethbridge", aliases = c("lethbridge")),
    list(eng = "Kelowna", fra = "Kelowna", aliases = c("kelowna")),
    list(
      eng = "Abbotsford - Mission", fra = "Abbotsford - Mission",
      aliases = c("abbotsford mission", "abbotsford")
    ),
    list(eng = "Vancouver", fra = "Vancouver", aliases = c("vancouver")),
    list(eng = "Victoria", fra = "Victoria", aliases = c("victoria")),
    list(eng = "Nanaimo", fra = "Nanaimo", aliases = c("nanaimo")),
    list(eng = "Chilliwack", fra = "Chilliwack", aliases = c("chilliwack")),
    list(eng = "Kamloops", fra = "Kamloops", aliases = c("kamloops"))
  )
}


extract_statcan_geographies <- function(normalized_query) {
  remaining <- paste0(" ", normalized_query, " ")
  found <- list()
  for (geography in statcan_geography_dictionary()) {
    aliases <- unique(normalize_statcan_search_text(geography$aliases))
    aliases <- aliases[order(nchar(aliases), decreasing = TRUE)]
    matched <- aliases[vapply(
      aliases,
      function(alias) grepl(
        paste0(" ", alias, " "),
        remaining,
        fixed = TRUE
      ),
      logical(1)
    )]
    if (!length(matched)) {
      next
    }
    found[[length(found) + 1L]] <- geography
    remaining <- gsub(
      paste0(" ", matched[[1L]], " "),
      " ",
      remaining,
      fixed = TRUE
    )
    remaining <- paste0(" ", trimws(gsub(" +", " ", remaining)), " ")
  }

  geographies <- if (length(found)) {
    data.frame(
      eng = vapply(found, `[[`, character(1), "eng"),
      fra = vapply(found, `[[`, character(1), "fra"),
      aliases = I(lapply(found, `[[`, "aliases")),
      stringsAsFactors = FALSE
    )
  } else {
    data.frame(
      eng = character(),
      fra = character(),
      aliases = I(list()),
      stringsAsFactors = FALSE
    )
  }

  list(
    geographies = geographies,
    remaining_text = trimws(gsub(" +", " ", remaining))
  )
}


extract_statcan_query_dates <- function(normalized_query, lang) {
  year_matches <- regmatches(
    normalized_query,
    gregexpr("\\b(?:19|20)[0-9]{2}\\b", normalized_query, perl = TRUE)
  )[[1L]]
  if (!length(year_matches) || identical(year_matches, "")) {
    return(list(
      start_date = as.Date(NA),
      end_date = as.Date(NA),
      label = NA_character_
    ))
  }

  years <- as.integer(year_matches)
  if (length(years) >= 2L) {
    start_year <- min(years)
    end_year <- max(years)
    label <- if (lang == "eng") {
      paste(start_year, "to", end_year)
    } else {
      paste(start_year, "\u00e0", end_year)
    }
    return(list(
      start_date = as.Date(paste0(start_year, "-01-01")),
      end_date = as.Date(paste0(end_year, "-12-31")),
      label = label
    ))
  }

  year <- years[[1L]]
  starts_at_year <- grepl(
    "\\b(since|from|after|starting|depuis)\\b|a partir",
    normalized_query,
    perl = TRUE
  )
  ends_at_year <- grepl(
    "\\b(until|through|before|jusqu)\\b",
    normalized_query,
    perl = TRUE
  )
  if (starts_at_year && !ends_at_year) {
    return(list(
      start_date = as.Date(paste0(year, "-01-01")),
      end_date = as.Date(NA),
      label = if (lang == "eng") {
        paste("since", year)
      } else {
        paste("depuis", year)
      }
    ))
  }
  if (ends_at_year && !starts_at_year) {
    return(list(
      start_date = as.Date(NA),
      end_date = as.Date(paste0(year, "-12-31")),
      label = if (lang == "eng") {
        paste("through", year)
      } else {
        paste("jusqu'en", year)
      }
    ))
  }

  list(
    start_date = as.Date(paste0(year, "-01-01")),
    end_date = as.Date(paste0(year, "-12-31")),
    label = if (lang == "eng") paste("year", year) else {
      paste("ann\u00e9e", year)
    }
  )
}


# Tokenizing every catalogue title is the dominant cost of statcan_find(), yet
# the tokens are a pure function of the title text and change only when the
# 24-hour catalogue cache is refreshed. Compute them once per catalogue and
# reuse them, both within the session (a package environment) and across
# sessions (a cache file beside the catalogue). Both layers are validated by
# comparing the stored titles to the current catalogue titles, so any catalogue
# change transparently invalidates stale tokens.
statcan_token_cache <- new.env(parent = emptyenv())

statcan_catalogue_tokens <- function(catalogue, lang) {
  title_column <- if (lang == "eng") "title_eng" else "title_fra"
  titles <- catalogue[[title_column]]
  # Tokens depend on the title text *and* on the tokenizer (stopwords,
  # synonyms, normalization). StatCan titles are stable for months, so keying
  # only on titles would silently serve tokens from an older tokenizer after an
  # upgrade. Stamping each entry with the package version fails safe: an
  # unrelated release triggers one harmless rebuild instead of stale rankings.
  version <- as.character(utils::packageVersion("statcanR"))

  memo <- statcan_token_cache[[lang]]
  if (is.list(memo) && identical(memo$version, version) &&
      identical(memo$titles, titles)) {
    return(memo$tokens)
  }

  cache_dir <- tools::R_user_dir("statcanR", which = "cache")
  cache_file <- file.path(cache_dir, "statcan_catalogue_tokens.rds")
  disk <- if (file.exists(cache_file)) {
    tryCatch(readRDS(cache_file), error = function(error) NULL)
  } else {
    NULL
  }
  entry <- if (is.list(disk)) disk[[lang]] else NULL
  if (is.list(entry) && identical(entry$version, version) &&
      is.character(entry$titles) && identical(entry$titles, titles) &&
      is.list(entry$tokens) && length(entry$tokens) == length(titles)) {
    statcan_token_cache[[lang]] <- entry
    return(entry$tokens)
  }

  tokens <- lapply(titles, canonical_statcan_tokens)
  entry <- list(version = version, titles = titles, tokens = tokens)
  statcan_token_cache[[lang]] <- entry

  updated <- if (is.list(disk)) disk else list()
  updated[[lang]] <- entry
  dir.create(cache_dir, recursive = TRUE, showWarnings = FALSE)
  tryCatch(
    saveRDS(updated, cache_file, version = 3L),
    error = function(error) NULL
  )

  tokens
}


rank_statcan_catalogue <- function(catalogue, parsed, lang) {
  title_column <- if (lang == "eng") "title_eng" else "title_fra"
  titles <- catalogue[[title_column]]
  title_terms <- statcan_catalogue_tokens(catalogue, lang)
  query_terms <- parsed$topic_terms

  matched_count <- statcan_matched_term_counts(title_terms, query_terms)
  match_ratio <- matched_count / length(query_terms)
  title_term_count <- vapply(
    title_terms,
    function(terms) max(1L, length(unique(terms))),
    integer(1)
  )
  # Prefer a concise title when two tables match the same query terms. This
  # tends to rank a broad table ahead of a highly specialized table whose
  # title happens to contain the same words.
  topic_density <- matched_count / title_term_count

  date_match <- statcan_date_coverage(
    catalogue$start_date,
    catalogue$end_date,
    parsed$start_date,
    parsed$end_date
  )
  keep <- matched_count > 0L & (is.na(date_match) | date_match)
  if (!any(keep)) {
    return(data.frame())
  }

  current_bonus <- ifelse(
    is.na(catalogue$archived),
    0,
    ifelse(catalogue$archived, -10, 5)
  )
  date_bonus <- ifelse(is.na(date_match), 0, ifelse(date_match, 5, 0))
  score <- round(
    60 * match_ratio +
      10 * (matched_count == length(query_terms)) +
      10 * topic_density +
      date_bonus +
      current_bonus,
    1
  )

  candidates <- data.frame(
    title = titles[keep],
    id = catalogue$id[keep],
    score = score[keep],
    start_date = catalogue$start_date[keep],
    end_date = catalogue$end_date[keep],
    release_date = catalogue$release_date[keep],
    archived = catalogue$archived[keep],
    matched_count = matched_count[keep],
    topic_count = rep(length(query_terms), sum(keep)),
    date_match = date_match[keep],
    stringsAsFactors = FALSE
  )
  order_statcan_candidates(candidates)
}


# For each title, count how many query terms it matches. A term matches on an
# exact token hit, or (for terms of five or more characters) when it shares a
# prefix with a title token of five or more characters. The per-title long-token
# filter and the query-term lengths are computed once, and the prefix test is
# vectorized over a title's tokens, so the work does not scale with
# titles * query_terms the way a naive nested loop would.
statcan_matched_term_counts <- function(title_terms, query_terms) {
  query_is_long <- nchar(query_terms) >= 5L
  n_query <- length(query_terms)
  vapply(
    title_terms,
    function(terms) {
      long_terms <- terms[nchar(terms) >= 5L]
      have_long <- length(long_terms) > 0L
      count <- 0L
      for (i in seq_len(n_query)) {
        term <- query_terms[[i]]
        if (term %in% terms) {
          count <- count + 1L
        } else if (query_is_long[[i]] && have_long &&
                   any(startsWith(long_terms, term) |
                       startsWith(term, long_terms))) {
          count <- count + 1L
        }
      }
      count
    },
    integer(1)
  )
}


statcan_date_coverage <- function(table_start, table_end,
                                  query_start, query_end) {
  if (is.na(query_start) && is.na(query_end)) {
    return(rep(NA, length(table_start)))
  }
  required_start <- if (is.na(query_start)) query_end else query_start
  required_end <- if (is.na(query_end)) query_start else query_end
  known <- !is.na(table_start) & !is.na(table_end)
  result <- rep(NA, length(table_start))
  result[known] <- table_start[known] <= required_start &
    table_end[known] >= required_end
  result
}


order_statcan_candidates <- function(candidates) {
  release_order <- as.numeric(candidates$release_date)
  release_order[is.na(release_order)] <- -Inf
  candidates[
    order(-candidates$score, -release_order, candidates$title),
    ,
    drop = FALSE
  ]
}


parsed_geography_label <- function(parsed, lang) {
  if (!nrow(parsed$geographies)) {
    return(NA_character_)
  }
  paste(parsed$geographies[[lang]], collapse = ", ")
}


build_statcan_match_reason <- function(candidate, parsed, lang) {
  if (lang == "eng") {
    reasons <- paste0(
      candidate$matched_count, "/", candidate$topic_count,
      " topic terms matched"
    )
    if (nrow(parsed$geographies)) {
      reasons <- c(
        reasons,
        if (candidate$geography_match %in% TRUE) {
          paste0("includes ", candidate$geography)
        } else {
          "geography could not be verified"
        }
      )
    }
    if (!is.na(parsed$date_label)) {
      reasons <- c(
        reasons,
        if (candidate$date_match %in% TRUE) {
          paste0("covers ", parsed$date_label)
        } else {
          "date coverage unavailable"
        }
      )
    }
    if (!is.na(candidate$archived)) {
      reasons <- c(
        reasons,
        if (candidate$archived) "archived table" else "current table"
      )
    }
  } else {
    reasons <- paste0(
      candidate$matched_count, "/", candidate$topic_count,
      " termes th\u00e9matiques correspondants"
    )
    if (nrow(parsed$geographies)) {
      reasons <- c(
        reasons,
        if (candidate$geography_match %in% TRUE) {
          paste0("comprend ", candidate$geography)
        } else {
          "g\u00e9ographie non v\u00e9rifi\u00e9e"
        }
      )
    }
    if (!is.na(parsed$date_label)) {
      reasons <- c(
        reasons,
        if (candidate$date_match %in% TRUE) {
          paste0("couvre ", parsed$date_label)
        } else {
          "couverture chronologique indisponible"
        }
      )
    }
    if (!is.na(candidate$archived)) {
      reasons <- c(
        reasons,
        if (candidate$archived) "tableau archiv\u00e9" else "tableau actuel"
      )
    }
  }
  paste(reasons, collapse = "; ")
}


metadata_matches_geographies <- function(metadata, geographies) {
  members <- normalize_statcan_search_text(c(
    metadata$geographies_eng,
    metadata$geographies_fra
  ))
  all(vapply(
    seq_len(nrow(geographies)),
    function(index) {
      aliases <- normalize_statcan_search_text(
        unlist(geographies$aliases[index], use.names = FALSE)
      )
      any(vapply(
        aliases,
        function(alias) any(
          members == alias |
            startsWith(members, paste0(alias, " ")) |
            startsWith(alias, paste0(members, " "))
        ),
        logical(1)
      ))
    },
    logical(1)
  ))
}


statcan_metadata_for_tables <- function(table_ids, refresh = FALSE) {
  product_ids <- unique(vapply(
    table_ids,
    normalize_product_id,
    character(1)
  ))
  result <- stats::setNames(vector("list", length(product_ids)), product_ids)
  stale <- stats::setNames(vector("list", length(product_ids)), product_ids)
  cache_dir <- tools::R_user_dir("statcanR", which = "cache")
  dir.create(cache_dir, recursive = TRUE, showWarnings = FALSE)
  clean_statcan_metadata_cache(cache_dir)
  cache_ttl <- 7 * 24 * 60 * 60

  for (product_id in product_ids) {
    cache_file <- statcan_metadata_cache_file(cache_dir, product_id)
    if (!file.exists(cache_file)) {
      next
    }
    cached <- tryCatch(readRDS(cache_file), error = function(error) NULL)
    if (!is.list(cached) || !is_valid_statcan_cube_metadata(cached$metadata)) {
      next
    }
    stale[product_id] <- list(cached$metadata)
    age <- as.numeric(difftime(
      Sys.time(), file.info(cache_file)$mtime, units = "secs"
    ))
    if (!refresh && !is.na(age) && age < cache_ttl) {
      result[product_id] <- list(cached$metadata)
    }
  }

  needed <- product_ids[vapply(result, is.null, logical(1))]
  errors <- character()
  if (length(needed)) {
    chunks <- split(needed, ceiling(seq_along(needed) / 10L))
    for (chunk in chunks) {
      fetched <- tryCatch(
        fetch_statcan_cube_metadata(chunk),
        error = function(error) error
      )
      if (inherits(fetched, "error")) {
        errors <- c(errors, conditionMessage(fetched))
        next
      }
      for (product_id in names(fetched)) {
        result[product_id] <- list(fetched[[product_id]])
        cache_file <- statcan_metadata_cache_file(cache_dir, product_id)
        tryCatch(
          saveRDS(
            list(metadata = fetched[[product_id]]),
            cache_file,
            version = 3L
          ),
          error = function(error) NULL
        )
      }
    }
  }

  missing <- product_ids[vapply(result, is.null, logical(1))]
  for (product_id in missing) {
    if (!is.null(stale[[product_id]])) {
      result[product_id] <- list(stale[[product_id]])
    }
  }
  still_missing <- product_ids[vapply(result, is.null, logical(1))]
  if (length(errors)) {
    warning(
      "Unable to refresh some Statistics Canada table metadata; cached ",
      "metadata was used where available. ",
      paste(unique(errors), collapse = " "),
      call. = FALSE
    )
  } else if (length(still_missing)) {
    warning(
      "Statistics Canada did not return metadata for some candidates; ",
      "their geography could not be verified.",
      call. = FALSE
    )
  }
  result
}


statcan_metadata_cache_file <- function(cache_dir, product_id) {
  file.path(cache_dir, paste0("metadata_", product_id, ".rds"))
}


clean_statcan_metadata_cache <- function(cache_dir) {
  files <- list.files(
    cache_dir,
    pattern = "^metadata_[0-9]{8}\\.rds$",
    full.names = TRUE
  )
  if (!length(files)) {
    return(invisible())
  }
  age <- as.numeric(difftime(
    Sys.time(), file.info(files)$mtime, units = "days"
  ))
  outdated <- files[!is.na(age) & age > 30]
  if (length(outdated)) {
    unlink(outdated, force = TRUE)
  }
  invisible()
}


fetch_statcan_cube_metadata <- function(product_ids) {
  endpoint <- paste0(
    "https://www150.statcan.gc.ca/",
    "t1/wds/rest/getCubeMetadata"
  )
  body <- lapply(product_ids, function(product_id) {
    list(productId = as.numeric(product_id))
  })
  response <- httr::POST(
    endpoint,
    body = body,
    encode = "json",
    httr::timeout(60),
    httr::user_agent(statcan_user_agent())
  )
  httr::stop_for_status(response)
  payload <- jsonlite::fromJSON(
    httr::content(response, as = "text", encoding = "UTF-8"),
    simplifyVector = FALSE
  )
  normalize_statcan_cube_metadata(payload)
}


normalize_statcan_cube_metadata <- function(payload) {
  if (!is.list(payload)) {
    return(list())
  }
  entries <- if (!is.null(payload$status) && !is.null(payload$object)) {
    list(payload)
  } else {
    payload
  }
  result <- list()
  for (entry in entries) {
    if (!is.list(entry) || !identical(entry$status, "SUCCESS") ||
        !is.list(entry$object)) {
      next
    }
    object <- entry$object
    product_id <- tryCatch(
      normalize_product_id(as.character(object$productId)),
      error = function(error) NA_character_
    )
    if (is.na(product_id)) {
      next
    }
    dimensions <- object$dimension
    if (is.null(dimensions)) {
      dimensions <- object$dimensions
    }
    if (is.null(dimensions)) {
      dimensions <- list()
    }
    geography_dimensions <- Filter(function(dimension) {
      if (!is.list(dimension)) {
        return(FALSE)
      }
      names <- normalize_statcan_search_text(c(
        dimension$dimensionNameEn,
        dimension$dimensionNameFr
      ))
      any(names %in% c("geography", "geographie"))
    }, dimensions)

    geographies_eng <- unlist(lapply(geography_dimensions, function(dimension) {
      members <- dimension$member
      if (is.null(members)) dimension$members else members
    }), recursive = FALSE, use.names = FALSE)
    geographies_fra <- geographies_eng
    geographies_eng <- unique(vapply(
      geographies_eng,
      function(member) metadata_member_name(member, "memberNameEn"),
      character(1)
    ))
    geographies_fra <- unique(vapply(
      geographies_fra,
      function(member) metadata_member_name(member, "memberNameFr"),
      character(1)
    ))
    geographies_eng <- geographies_eng[nzchar(geographies_eng)]
    geographies_fra <- geographies_fra[nzchar(geographies_fra)]

    archive_value <- object$archiveStatusCode
    if (is.null(archive_value)) {
      archive_value <- object$archived
    }
    archived <- if (is.null(archive_value)) {
      NA
    } else {
      normalize_archive_status(archive_value)[[1L]]
    }
    metadata <- list(
      id = product_id,
      title_eng = decode_statcan_entities(
        metadata_scalar_character(object$cubeTitleEn)
      ),
      title_fra = decode_statcan_entities(
        metadata_scalar_character(object$cubeTitleFr)
      ),
      start_date = metadata_scalar_date(object$cubeStartDate),
      end_date = metadata_scalar_date(object$cubeEndDate),
      archived = archived,
      geographies_eng = geographies_eng,
      geographies_fra = geographies_fra
    )
    if (is_valid_statcan_cube_metadata(metadata)) {
      result[product_id] <- list(metadata)
    }
  }
  result
}


metadata_member_name <- function(member, field) {
  if (!is.list(member) || is.null(member[[field]]) ||
      length(member[[field]]) != 1L || is.na(member[[field]])) {
    return("")
  }
  as.character(member[[field]])
}


metadata_scalar_character <- function(value) {
  if (is.null(value) || length(value) != 1L || is.na(value)) "" else {
    as.character(value)
  }
}


metadata_scalar_date <- function(value) {
  if (is.null(value) || length(value) != 1L || is.na(value)) {
    return(as.Date(NA))
  }
  suppressWarnings(as.Date(substr(as.character(value), 1L, 10L)))
}


is_valid_statcan_cube_metadata <- function(metadata) {
  is.list(metadata) &&
    is.character(metadata$id) && length(metadata$id) == 1L &&
    grepl("^[0-9]{8}$", metadata$id) &&
    is.character(metadata$title_eng) && length(metadata$title_eng) == 1L &&
    is.character(metadata$title_fra) && length(metadata$title_fra) == 1L &&
    inherits(metadata$start_date, "Date") &&
    length(metadata$start_date) == 1L &&
    inherits(metadata$end_date, "Date") &&
    length(metadata$end_date) == 1L &&
    is.logical(metadata$archived) && length(metadata$archived) == 1L &&
    is.character(metadata$geographies_eng) &&
    is.character(metadata$geographies_fra)
}

test_that("natural-language queries identify topic, geography, and date", {
  parsed <- statcanR:::parse_statcan_query(
    "R&D expenditures in Quebec since 2020",
    "eng"
  )

  expect_identical(
    parsed$topic_terms,
    c("research", "development", "expenditure")
  )
  expect_identical(parsed$geographies$eng, "Quebec")
  expect_identical(parsed$start_date, as.Date("2020-01-01"))
  expect_true(is.na(parsed$end_date))
  expect_identical(parsed$date_label, "since 2020")
})

test_that("French natural-language queries are normalized", {
  parsed <- statcanR:::parse_statcan_query(
    "Dépenses de R-D au Québec depuis 2020",
    "fra"
  )

  expect_identical(
    parsed$topic_terms,
    c("expenditure", "research", "development")
  )
  expect_identical(parsed$geographies$fra, "Québec")
  expect_identical(parsed$start_date, as.Date("2020-01-01"))
  expect_identical(parsed$date_label, "depuis 2020")
})

test_that("catalogue candidates are ranked by topic and date coverage", {
  parsed <- statcanR:::parse_statcan_query(
    "R&D expenditures since 2020",
    "eng"
  )
  result <- statcanR:::rank_statcan_catalogue(
    sample_catalogue(),
    parsed,
    "eng"
  )

  expect_identical(result$id[[1L]], "27-10-0273-01")
  expect_true(result$matched_count[[1L]] == 3L)
  expect_true(result$date_match[[1L]])
})

test_that("catalogue tokens are cached and invalidated when titles change", {
  catalogue <- sample_catalogue()

  first <- statcanR:::statcan_catalogue_tokens(catalogue, "eng")
  expect_identical(
    first,
    lapply(catalogue$title_eng, statcanR:::canonical_statcan_tokens)
  )

  cache_file <- file.path(
    tools::R_user_dir("statcanR", which = "cache"),
    "statcan_catalogue_tokens.rds"
  )
  expect_true(file.exists(cache_file))

  # A second call reuses the cache and returns the same tokens.
  expect_identical(
    statcanR:::statcan_catalogue_tokens(catalogue, "eng"),
    first
  )

  # Changing a title must invalidate the stale tokens for that row.
  changed <- catalogue
  changed$title_eng[[1L]] <- "Household spending on tomatoes"
  refreshed <- statcanR:::statcan_catalogue_tokens(changed, "eng")
  expect_identical(
    refreshed[[1L]],
    statcanR:::canonical_statcan_tokens(changed$title_eng[[1L]])
  )
  expect_false(identical(refreshed[[1L]], first[[1L]]))

  # A cache written by a different package version must be rebuilt, even when
  # the titles are unchanged (the tokenizer itself may have changed).
  rm(list = ls(statcanR:::statcan_token_cache), envir = statcanR:::statcan_token_cache)
  saveRDS(
    list(eng = list(
      version = "0.0.0",
      titles = catalogue$title_eng,
      tokens = rep(list("stale"), nrow(catalogue))
    )),
    cache_file,
    version = 3L
  )
  expect_identical(
    statcanR:::statcan_catalogue_tokens(catalogue, "eng"),
    lapply(catalogue$title_eng, statcanR:::canonical_statcan_tokens)
  )
})

test_that("cube metadata exposes geography members", {
  payload <- list(list(
    status = "SUCCESS",
    object = list(
      productId = "27100273",
      cubeTitleEn = "Gross domestic expenditures on research and development",
      cubeTitleFr = paste(
        "Dépenses intérieures brutes en recherche et développement"
      ),
      cubeStartDate = "1963-01-01",
      cubeEndDate = "2025-01-01",
      archiveStatusCode = "2",
      dimension = list(list(
        dimensionNameEn = "Geography",
        dimensionNameFr = "Géographie",
        member = list(
          list(memberNameEn = "Canada", memberNameFr = "Canada"),
          list(memberNameEn = "Quebec", memberNameFr = "Québec")
        )
      ))
    )
  ))

  result <- statcanR:::normalize_statcan_cube_metadata(payload)

  expect_named(result, "27100273")
  expect_identical(
    result[["27100273"]]$geographies_eng,
    c("Canada", "Quebec")
  )
  expect_false(result[["27100273"]]$archived)
  expect_true(statcanR:::is_valid_statcan_cube_metadata(
    result[["27100273"]]
  ))
})

test_that("statcan_find returns verified ranked candidates", {
  local_mocked_bindings(
    statcan_catalogue = function(refresh = FALSE) sample_catalogue(),
    statcan_metadata_for_tables = function(table_ids, refresh = FALSE) {
      list(
        `27100273` = list(
          id = "27100273",
          title_eng = "Gross domestic expenditures on research and development",
          title_fra = paste(
            "Dépenses intérieures brutes en recherche et développement"
          ),
          start_date = as.Date("1963-01-01"),
          end_date = as.Date("2025-01-01"),
          archived = FALSE,
          geographies_eng = c("Canada", "Quebec"),
          geographies_fra = c("Canada", "Québec")
        ),
        `27100015` = list(
          id = "27100015",
          title_eng = "Research and development personnel",
          title_fra = "Personnel affecté à la recherche et au développement",
          start_date = as.Date("2015-01-01"),
          end_date = as.Date("2025-01-01"),
          archived = FALSE,
          geographies_eng = "Canada",
          geographies_fra = "Canada"
        )
      )
    },
    .package = "statcanR"
  )

  result <- statcan_find(
    "R&D expenditures in Quebec since 2020",
    n = 5
  )

  expect_s3_class(result, "data.frame")
  expect_identical(result$id, "27-10-0273-01")
  expect_true(result$geography_match)
  expect_match(result$match_reason, "includes Quebec")
  expect_identical(result$rank, 1L)
})

test_that("statcan_find does not request metadata without a geography", {
  local_mocked_bindings(
    statcan_catalogue = function(refresh = FALSE) sample_catalogue(),
    statcan_metadata_for_tables = function(...) {
      stop("metadata should not be requested")
    },
    .package = "statcanR"
  )

  result <- statcan_find("international exports", n = 1)

  expect_identical(result$id, "36-10-0434-01")
  expect_true(is.na(result$geography_match))
})

test_that("statcan_find validates query and result count", {
  expect_error(statcan_find("Quebec since 2020"), "include a subject")
  expect_error(statcan_find("population", n = 0), "1 to 20")
  expect_error(statcan_find(character()), "one non-empty")
})

#' Download a Statistics Canada data table
#'
#' Downloads a complete table from Statistics Canada's Web Data Service (WDS)
#' and returns it as a data frame. Product IDs can be supplied in the familiar
#' hyphenated form (for example, `"27-10-0014-01"`) or as an eight-digit PID
#' (for example, `"27100014"`).
#'
#' The function keeps the interface used by earlier statcanR releases. English
#' and French downloads share the same processing rules. In particular, the
#' first column is named `REF_DATE`, coordinates are stored as character, and
#' the table title from the metadata file is added as `INDICATOR`.
#'
#' @param tableNumber A Statistics Canada table number or Product ID. Both
#'   `"27-10-0014-01"` and `"27100014"` are accepted.
#' @param lang Language of the downloaded table: `"eng"` or `"fra"`.
#'
#' @return A data frame containing the complete Statistics Canada table.
#' @export
#'
#' @examples
#' \dontrun{
#' science <- statcan_data("27-10-0014-01", "eng")
#' science_fr <- statcan_data("27100014", "fra")
#' }
statcan_data <- function(tableNumber, lang) {
  product_id <- normalize_product_id(tableNumber)
  lang <- normalize_language(lang)

  work_dir <- tempfile("statcanR-")
  if (!dir.create(work_dir, recursive = TRUE)) {
    stop("Unable to create a temporary directory for the download.",
         call. = FALSE)
  }
  on.exit(unlink(work_dir, recursive = TRUE, force = TRUE), add = TRUE)

  download_url <- statcan_download_url(product_id, lang)
  zip_file <- file.path(work_dir, paste0(product_id, ".zip"))

  message("statcanR: downloading table ", format_product_id(product_id), ".")
  response <- tryCatch(
    httr::GET(
      download_url,
      httr::timeout(300),
      httr::user_agent(statcan_user_agent()),
      httr::write_disk(zip_file, overwrite = TRUE)
    ),
    error = function(error) {
      stop(
        "Unable to download Statistics Canada table ",
        format_product_id(product_id), ". ", conditionMessage(error),
        call. = FALSE
      )
    }
  )
  stop_for_statcan_status(response, product_id)

  read_statcan_zip(zip_file, product_id, lang, work_dir)
}


#' Download a Statistics Canada table and save a CSV file
#'
#' Calls [statcan_data()] and also writes the returned table to a CSV file.
#' Existing calls with only `tableNumber` and `lang` remain supported; `path`
#' can be used to select a different output directory.
#'
#' @inheritParams statcan_data
#' @param path Directory in which to save the CSV file. The directory must
#'   already exist. Defaults to the current working directory.
#'
#' @return The downloaded table. The CSV path is available in the
#'   `statcan_file` attribute of the returned data frame.
#' @export
#'
#' @examples
#' \dontrun{
#' science <- statcan_download_data("27-10-0014-01", "eng")
#' science <- statcan_download_data(
#'   "27-10-0014-01", "eng", path = tempdir()
#' )
#' attr(science, "statcan_file")
#' }
statcan_download_data <- function(tableNumber, lang, path = ".") {
  product_id <- normalize_product_id(tableNumber)
  lang <- normalize_language(lang)
  path <- normalize_output_path(path)

  can_data <- statcan_data(product_id, lang)
  output_file <- file.path(
    path,
    paste0("statcan_", product_id, "_", lang, ".csv")
  )

  utils::write.csv(
    can_data,
    file = output_file,
    row.names = FALSE,
    na = "",
    fileEncoding = "UTF-8"
  )
  attr(can_data, "statcan_file") <- normalizePath(
    output_file,
    winslash = "/",
    mustWork = FALSE
  )
  message("statcanR: saved ", output_file, ".")

  can_data
}


# Convert supported table-number formats to the eight-digit WDS PID.
normalize_product_id <- function(table_number) {
  if (!is.character(table_number) || length(table_number) != 1L ||
      is.na(table_number) || !nzchar(trimws(table_number))) {
    stop("`tableNumber` must be one non-empty character string.",
         call. = FALSE)
  }

  product_id <- gsub("[^0-9]", "", trimws(table_number))
  if (nchar(product_id) == 10L) {
    product_id <- substr(product_id, 1L, 8L)
  }
  if (!grepl("^[0-9]{8}$", product_id)) {
    stop(
      "`tableNumber` must contain an eight-digit Product ID, optionally ",
      "followed by a two-digit view (for example, '27-10-0014-01').",
      call. = FALSE
    )
  }

  product_id
}


# Preserve the public eng/fra values while accepting the WDS aliases too.
normalize_language <- function(lang) {
  if (!is.character(lang) || length(lang) != 1L || is.na(lang)) {
    stop("`lang` must be one of 'eng' or 'fra'.", call. = FALSE)
  }

  aliases <- c(eng = "eng", en = "eng", fra = "fra", fr = "fra")
  normalized <- unname(aliases[tolower(trimws(lang))])
  if (length(normalized) != 1L || is.na(normalized)) {
    stop("`lang` must be one of 'eng' or 'fra'.", call. = FALSE)
  }

  normalized
}


normalize_output_path <- function(path) {
  if (!is.character(path) || length(path) != 1L || is.na(path) ||
      !nzchar(path)) {
    stop("`path` must be one existing directory.", call. = FALSE)
  }
  if (!dir.exists(path)) {
    stop("Output directory does not exist: ", path, call. = FALSE)
  }

  path
}


format_product_id <- function(product_id) {
  paste0(
    substr(product_id, 1L, 2L), "-",
    substr(product_id, 3L, 4L), "-",
    substr(product_id, 5L, 8L), "-01"
  )
}


statcan_user_agent <- function() {
  version <- tryCatch(
    as.character(utils::packageVersion("statcanR")),
    error = function(error) "development"
  )
  paste0(
    "statcanR/", version,
    " (https://github.com/warint/statcanR)"
  )
}


# Ask WDS for the current full-table download URL.
statcan_download_url <- function(product_id, lang) {
  wds_language <- if (lang == "eng") "en" else "fr"
  endpoint <- paste0(
    "https://www150.statcan.gc.ca/t1/wds/rest/",
    "getFullTableDownloadCSV/", product_id, "/", wds_language
  )

  response <- tryCatch(
    httr::GET(
      endpoint,
      httr::timeout(60),
      httr::user_agent(statcan_user_agent())
    ),
    error = function(error) {
      stop(
        "Unable to contact Statistics Canada's Web Data Service for table ",
        format_product_id(product_id), ". ", conditionMessage(error),
        call. = FALSE
      )
    }
  )
  stop_for_statcan_status(response, product_id)

  payload <- tryCatch(
    jsonlite::fromJSON(
      httr::content(response, as = "text", encoding = "UTF-8"),
      simplifyVector = TRUE
    ),
    error = function(error) {
      stop(
        "Statistics Canada's Web Data Service returned invalid JSON. ",
        conditionMessage(error),
        call. = FALSE
      )
    }
  )

  if (!is.list(payload) || !identical(payload$status, "SUCCESS") ||
      !is.character(payload$object) || length(payload$object) != 1L ||
      !grepl("^https://", payload$object)) {
    details <- if (is.list(payload) && !is.null(payload$object)) {
      paste(as.character(payload$object), collapse = " ")
    } else {
      "unexpected response"
    }
    stop(
      "Statistics Canada did not provide a download for table ",
      format_product_id(product_id), ": ", details, ".",
      call. = FALSE
    )
  }

  payload$object
}


stop_for_statcan_status <- function(response, product_id) {
  if (!httr::http_error(response)) {
    return(invisible(response))
  }

  status <- httr::status_code(response)
  detail <- tryCatch(
    httr::content(response, as = "text", encoding = "UTF-8"),
    error = function(error) ""
  )
  detail <- trimws(gsub("[\r\n]+", " ", detail))
  if (nchar(detail) > 200L) {
    detail <- paste0(substr(detail, 1L, 200L), "...")
  }
  suffix <- if (nzchar(detail)) paste0(" ", detail) else ""

  stop(
    "Statistics Canada returned HTTP ", status, " for table ",
    format_product_id(product_id), ".", suffix,
    call. = FALSE
  )
}


read_statcan_zip <- function(zip_file, product_id, lang, work_dir) {
  archive <- tryCatch(
    utils::unzip(zip_file, list = TRUE),
    error = function(error) {
      stop("The Statistics Canada download is not a valid ZIP archive. ",
           conditionMessage(error), call. = FALSE)
    }
  )
  archive_names <- archive$Name
  data_name <- archive_names[
    basename(archive_names) == paste0(product_id, ".csv")
  ]
  metadata_name <- archive_names[
    basename(archive_names) == paste0(product_id, "_MetaData.csv")
  ]

  if (length(data_name) != 1L || length(metadata_name) != 1L) {
    stop(
      "The Statistics Canada ZIP archive does not contain the expected data ",
      "and metadata files.",
      call. = FALSE
    )
  }

  extracted <- tryCatch(
    utils::unzip(
      zip_file,
      files = c(data_name, metadata_name),
      exdir = work_dir
    ),
    error = function(error) {
      stop("Unable to extract the Statistics Canada ZIP archive. ",
           conditionMessage(error), call. = FALSE)
    }
  )
  data_file <- extracted[basename(extracted) == paste0(product_id, ".csv")]
  metadata_file <- extracted[
    basename(extracted) == paste0(product_id, "_MetaData.csv")
  ]

  can_data <- data.table::fread(
    data_file,
    encoding = "UTF-8",
    showProgress = FALSE
  )
  metadata <- read_statcan_metadata(metadata_file)
  if (!ncol(can_data) || !nrow(metadata) || !ncol(metadata)) {
    stop("Statistics Canada returned an empty data or metadata file.",
         call. = FALSE)
  }

  data.table::setnames(can_data, 1L, "REF_DATE")
  reference_dates <- parse_reference_dates(can_data[["REF_DATE"]])
  has_fiscal <- isTRUE(attr(reference_dates, "statcan_has_fiscal"))
  attr(reference_dates, "statcan_has_fiscal") <- NULL
  can_data[["REF_DATE"]] <- reference_dates
  coordinate_columns <- intersect(
    c("COORDINATE", "COORDONN\u00c9ES", "COORDONNEES"),
    names(can_data)
  )
  for (coordinate_column in coordinate_columns) {
    can_data[[coordinate_column]] <- as.character(can_data[[coordinate_column]])
  }
  can_data <- stabilize_statcan_empty_columns(can_data)
  can_data[["INDICATOR"]] <- as.character(metadata[[1L]][1L])

  if (has_fiscal) {
    can_data[["REF_PERIOD"]] <- if (lang == "eng") {
      "Fiscal year"
    } else {
      "Exercice financier"
    }
    data.table::setcolorder(
      can_data,
      c("REF_DATE", "REF_PERIOD", setdiff(names(can_data),
                                            c("REF_DATE", "REF_PERIOD")))
    )
  }

  data.table::setDF(can_data)
  can_data
}


# Give columns that are blank throughout a table a stable type.
#
# data.table::fread() types a column that is empty for every row as logical
# (all NA). Statistics Canada tables contain no boolean columns, so such a
# column is really an unpopulated text column - DGUID, STATUS, SYMBOL and
# TERMINATED are commonly empty. Left as logical, a column's type would depend
# on whether the particular table happened to populate it, which breaks code
# that binds several tables together or expects, say, STATUS to be character.
# Converting only all-NA logical columns is safe: a column that carries real
# flags (for example "F" or "t") is already character and is left untouched.
stabilize_statcan_empty_columns <- function(can_data) {
  for (name in names(can_data)) {
    column <- can_data[[name]]
    if (is.logical(column) && all(is.na(column))) {
      can_data[[name]] <- as.character(column)
    }
  }
  can_data
}


# Read the cube-level row from a Statistics Canada "_MetaData.csv" file.
#
# These files are not a single rectangular table: a header and one cube row are
# followed by a blank line and further sections (dimensions, members, ...) that
# have different column counts. data.table::fread() samples the whole file to
# detect its shape, so the ragged later sections can make it stop early and
# return nothing even with nrows = 1 (see issue #8). We only need the first
# section, so we hand fread() just the header and the cube row.
read_statcan_metadata <- function(metadata_file) {
  lines <- tryCatch(
    readLines(metadata_file, encoding = "UTF-8", warn = FALSE),
    error = function(error) character()
  )

  # Sections are separated by a blank line, so the header and cube row are
  # everything before the first blank line. Slicing here (rather than reading a
  # fixed two lines) keeps the cube row intact even if a title ever spans more
  # than one physical line.
  blank <- which(!nzchar(trimws(lines)))
  first_section <- if (length(blank)) {
    lines[seq_len(blank[[1L]] - 1L)]
  } else {
    lines
  }
  if (length(first_section) < 2L) {
    return(data.table::data.table())
  }

  tryCatch(
    data.table::fread(
      text = first_section,
      fill = TRUE,
      encoding = "UTF-8",
      showProgress = FALSE
    ),
    error = function(error) data.table::data.table()
  )
}


parse_reference_dates <- function(reference_date) {
  reference_date <- as.character(reference_date)
  result <- rep(as.Date(NA), length(reference_date))

  fiscal <- grepl("^[0-9]{4}/[0-9]{4}$", reference_date)
  year_month <- grepl("^[0-9]{4}-[0-9]{2}$", reference_date)
  year <- grepl("^[0-9]{4}$", reference_date)
  iso_date <- grepl("^[0-9]{4}-[0-9]{2}-[0-9]{2}$", reference_date)

  if (any(fiscal)) {
    result[fiscal] <- as.Date(paste0(
      sub(".*/", "", reference_date[fiscal]),
      "-03-31"
    ))
  }
  if (any(year_month)) {
    result[year_month] <- as.Date(paste0(
      reference_date[year_month],
      "-01"
    ))
  }
  if (any(year)) {
    result[year] <- as.Date(paste0(reference_date[year], "-01-01"))
  }
  if (any(iso_date)) {
    result[iso_date] <- as.Date(reference_date[iso_date])
  }

  unsupported <- !is.na(reference_date) &
    !(fiscal | year_month | year | iso_date)
  if (any(unsupported)) {
    warning(
      "Some REF_DATE values use an unsupported format and were converted to NA.",
      call. = FALSE
    )
  }

  # Expose whether any fiscal-year values were seen so the caller need not
  # rescan the whole reference-date column to decide on the REF_PERIOD label.
  attr(result, "statcan_has_fiscal") <- any(fiscal)
  result
}

test_that("Product IDs retain compatibility with earlier input formats", {
  expect_identical(
    statcanR:::normalize_product_id("27-10-0014-01"),
    "27100014"
  )
  expect_identical(statcanR:::normalize_product_id("27100014"), "27100014")
  expect_identical(
    statcanR:::normalize_product_id("27-10-0014"),
    "27100014"
  )
  expect_identical(
    statcanR:::format_product_id("27100014"),
    "27-10-0014-01"
  )
})

test_that("invalid Product IDs and languages fail clearly", {
  expect_error(statcanR:::normalize_product_id(27100014), "character")
  expect_error(statcanR:::normalize_product_id("27-10"), "eight-digit")
  expect_identical(statcanR:::normalize_language("eng"), "eng")
  expect_identical(statcanR:::normalize_language("fr"), "fra")
  expect_error(statcanR:::normalize_language("deu"), "eng.*fra")
})

test_that("reference periods are converted to stable dates", {
  dates <- statcanR:::parse_reference_dates(
    c("2020/2021", "2024-07", "2025", "2026-07-17")
  )

  expect_s3_class(dates, "Date")
  expect_identical(
    format(dates),
    c("2021-03-31", "2024-07-01", "2025-01-01", "2026-07-17")
  )

  annual <- statcanR:::parse_reference_dates(1999:2001)
  expect_identical(
    format(annual),
    c("1999-01-01", "2000-01-01", "2001-01-01")
  )
})

test_that("multi-section metadata files yield the cube title (issue #8)", {
  metadata_file <- tempfile(fileext = ".csv")
  on.exit(unlink(metadata_file), add = TRUE)

  # Mirror the real "_MetaData.csv" layout: a header and one cube row, then a
  # blank line and further sections with different column counts. The cube row
  # has a comma-bearing quoted title and an unquoted, semicolon-separated field.
  writeLines(
    c(
      paste0(
        '"Cube Title","Product Id","CANSIM Id",URL,"Cube Notes",',
        '"Archive Status",Frequency,"Start Reference Period",',
        '"End Reference Period","Total number of dimensions"'
      ),
      paste0(
        '"Life expectancy, single-year estimates, Canada, all provinces ',
        'except Prince Edward Island","13100837",,',
        '"https://www150.statcan.gc.ca/t1/tbl1/en/tv.action?pid=1310083701",',
        '1;2;3;13,"CURRENT - a cube available to the public and that is ',
        'current","Annual","1980-01-01","2024-01-01","4",'
      ),
      "",
      '"Dimension ID","Dimension name","Dimension Notes","Dimension Definitions"',
      '1,"Geography",,',
      '2,"Age group",,'
    ),
    metadata_file,
    useBytes = TRUE
  )

  metadata <- statcanR:::read_statcan_metadata(metadata_file)

  # Exactly the cube row: if later sections bled in, nrow would exceed 1.
  expect_identical(nrow(metadata), 1L)
  expect_identical(
    as.character(metadata[[1L]][1L]),
    paste0(
      "Life expectancy, single-year estimates, Canada, all provinces ",
      "except Prince Edward Island"
    )
  )
})

test_that("empty or truncated metadata files return an empty table", {
  short_file <- tempfile(fileext = ".csv")
  on.exit(unlink(short_file), add = TRUE)
  writeLines('"Cube Title","Product Id"', short_file)

  metadata <- statcanR:::read_statcan_metadata(short_file)
  expect_s3_class(metadata, "data.frame")
  expect_identical(nrow(metadata), 0L)
})

test_that("empty Statistics Canada columns get a stable character type", {
  data_csv <- tempfile(fileext = ".csv")
  on.exit(unlink(data_csv), add = TRUE)
  # DGUID, STATUS, SYMBOL and TERMINATED are blank for every row, as in many
  # real tables; the data-bearing columns are populated.
  writeLines(
    c(
      paste(
        "REF_DATE", "GEO", "DGUID", "Sector", "UOM", "UOM_ID",
        "SCALAR_FACTOR", "SCALAR_ID", "VECTOR", "COORDINATE", "VALUE",
        "STATUS", "SYMBOL", "TERMINATED", "DECIMALS",
        sep = ","
      ),
      "2020,Canada,,Public sector,Persons,249,units,0,v1,1.1,100,,,,0",
      "2021,Canada,,Public sector,Persons,249,units,0,v1,1.1,110,,,,0"
    ),
    data_csv
  )
  raw <- data.table::fread(data_csv, encoding = "UTF-8", showProgress = FALSE)

  # Precondition: fread types the all-blank columns as logical.
  expect_true(is.logical(raw$DGUID))
  expect_true(is.logical(raw$STATUS))
  expect_true(is.logical(raw$SYMBOL))
  expect_true(is.logical(raw$TERMINATED))

  stable <- statcanR:::stabilize_statcan_empty_columns(raw)

  # The empty text columns are now character, still holding no values.
  for (nm in c("DGUID", "STATUS", "SYMBOL", "TERMINATED")) {
    expect_type(stable[[nm]], "character")
    expect_true(all(is.na(stable[[nm]])))
  }

  # Populated columns keep their natural type and values.
  expect_type(stable$GEO, "character")
  expect_identical(stable$GEO, c("Canada", "Canada"))
  expect_true(is.numeric(stable$VALUE))
  expect_equal(as.numeric(stable$VALUE), c(100, 110))
  expect_true(is.numeric(stable$UOM_ID))
})

test_that("type stabilization never rewrites populated flag columns", {
  data_csv <- tempfile(fileext = ".csv")
  on.exit(unlink(data_csv), add = TRUE)
  # STATUS carries reliability letters ("E", "F"); TERMINATED carries "t". The
  # "F" and "t" must survive as themselves, not be coerced to "FALSE"/logical.
  writeLines(
    c(
      "REF_DATE,GEO,VALUE,STATUS,TERMINATED",
      "2020,Canada,100,E,",
      "2021,Canada,,F,t"
    ),
    data_csv
  )
  raw <- data.table::fread(data_csv, encoding = "UTF-8", showProgress = FALSE)
  stable <- statcanR:::stabilize_statcan_empty_columns(raw)

  expect_type(stable$STATUS, "character")
  expect_identical(stable$STATUS, c("E", "F"))
  expect_identical(stable$TERMINATED, c("", "t"))
})

test_that("type stabilization is column-name agnostic", {
  data_csv <- tempfile(fileext = ".csv")
  on.exit(unlink(data_csv), add = TRUE)
  # An arbitrary, non-standard empty column stands in for a French header or a
  # column Statistics Canada might add later: the rule is driven by type, not
  # by a hard-coded list of names.
  writeLines(
    c(
      "REF_DATE,GEO,SOME_NEW_COLUMN,VALUE",
      "2020,Canada,,100",
      "2021,Canada,,110"
    ),
    data_csv
  )
  raw <- data.table::fread(data_csv, encoding = "UTF-8", showProgress = FALSE)
  expect_true(is.logical(raw$SOME_NEW_COLUMN))

  stable <- statcanR:::stabilize_statcan_empty_columns(raw)
  expect_type(stable$SOME_NEW_COLUMN, "character")
  expect_true(all(is.na(stable$SOME_NEW_COLUMN)))
})

test_that("download output paths must already exist", {
  expect_identical(statcanR:::normalize_output_path(tempdir()), tempdir())
  expect_error(
    statcanR:::normalize_output_path(file.path(tempdir(), "not-created")),
    "does not exist"
  )
})

test_that("statcan_download_data writes and returns its data", {
  output_dir <- tempfile("statcanR-test-")
  dir.create(output_dir)
  on.exit(unlink(output_dir, recursive = TRUE, force = TRUE), add = TRUE)

  local_mocked_bindings(
    statcan_data = function(tableNumber, lang) {
      data.frame(
        REF_DATE = as.Date("2026-01-01"),
        COORDINATE = "1.1",
        VALUE = 42,
        stringsAsFactors = FALSE
      )
    },
    .package = "statcanR"
  )

  expect_message(
    result <- statcan_download_data("10-10-0001-01", "eng", output_dir),
    "saved"
  )
  expected_file <- file.path(output_dir, "statcan_10100001_eng.csv")

  expect_s3_class(result, "data.frame")
  expect_true(file.exists(expected_file))
  expect_identical(
    attr(result, "statcan_file"),
    normalizePath(expected_file, winslash = "/", mustWork = TRUE)
  )
})

---
title: "Getting started with statcanR"
author: "Thierry Warin"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Getting started with statcanR}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = FALSE
)
```

## Overview

`statcanR` connects R to [Statistics Canada's Web Data Service (WDS)](https://www.statcan.gc.ca/en/developers/wds). This vignette walks through a complete workflow: describing the data you need, choosing a table, understanding its identifier, downloading it, inspecting the result, and optionally saving a CSV copy.

The package supports four common tasks:

1. Find likely tables from an ordinary-language description.
2. Search for exact keywords in the official table catalogue.
3. Download a complete table in English or French into R.
4. Save the downloaded table as a CSV file.

The four public functions have distinct purposes:

| Function | Use it when... | Result |
|---|---|---|
| `statcan_find()` | You can describe the subject, geography, or period you need | A ranked data frame of likely tables and reasons for each match |
| `statcan_search()` | You know words that occur in the official table title | An interactive table of exact keyword matches |
| `statcan_data()` | You want the complete table in R | A data frame |
| `statcan_download_data()` | You want the data frame and a CSV copy | A data frame with the saved file path attached |

The download functions retrieve a **complete table**, not a filtered selection of observations. A Statistics Canada table can be large. It is therefore useful to identify the right table before starting the download.

An optional fifth function, `statcan_chat()`, wraps `statcan_find()` with a user-configured language model to explain candidates and ask clarifying questions; see [Optional: ask a language model for help](#optional-ask-a-language-model-for-help).

## Two concepts to know first

### Table number and Product ID

Statistics Canada displays table numbers such as `10-10-0001-01`. The first eight digits form the WDS Product ID (PID), `10100001`; the final `01` identifies the displayed view. `statcanR` accepts either the displayed table number or the eight-digit PID:

```{r id-equivalence}
table_data <- statcan_data("10-10-0001-01", "eng")
table_data <- statcan_data("10100001", "eng")
```

These two calls request the same table.

### Language

Use `lang = "eng"` for an English table and `lang = "fra"` for a French table. The language controls the table contents and labels returned by Statistics Canada; it is not a translation performed by `statcanR`.

## Install or upgrade

The command used for a first installation also upgrades an older CRAN installation:

```{r install}
install.packages("statcanR")
```

Then load the package:

```{r setup}
library(statcanR)
```

If the package was loaded while you upgraded it, restart the R session before calling `library(statcanR)` again. Check which version R will use with:

```{r version}
packageVersion("statcanR")
```

Version 0.3.0 preserves the familiar calls from earlier releases. In particular, code that supplies a table number and a language to `statcan_data()` or `statcan_download_data()` continues to work.

## Step 1: describe the table you need

Use `statcan_find()` when you know what data you want but do not know the
official title or table number. Write a short request containing as much of the
subject, Canadian geography, and period as you know:

```{r find-eng}
matches <- statcan_find(
  "R&D expenditures in Quebec since 2020",
  lang = "eng",
  n = 5
)

matches[, c("title", "id", "score", "match_reason")]
```

This request is interpreted as:

- a **subject**: research and development expenditures;
- a **geography**: Quebec; and
- a **date requirement**: the table must include 2020.

The result is an ordinary data frame. `rank` orders the candidates, `score`
summarizes the strength of each match, and `match_reason` states which clues
were confirmed. The score helps with discovery; it is not a measure of data
quality. Read the candidate titles because two tables can represent different
valid interpretations of the same short request.

When a province or territory appears in the query, `statcan_find()` checks WDS
metadata to confirm that the table includes it. `start_date` and `end_date`
describe the coverage of the **whole table**. These checks help choose a table;
they do not select rows. The download functions still retrieve the complete
table, after which you filter its geography and reference-period columns.

French requests and French catalogue titles are supported too:

```{r find-fra}
matches_fr <- statcan_find(
  "Dépenses de R-D au Québec depuis 2020",
  lang = "fra"
)
```

The parser recognizes common wording and abbreviations, but it is deliberately
simple and predictable. A concise request generally works better than a long
question. If the results are too broad, add a more specific subject term. If
there are no results, remove a detail or try an official keyword.

### Search exact title keywords

Use `statcan_search()` when you know words that occur in the official title. It searches titles without regard to letter case:

```{r search-eng}
statcan_search(
  c("federal", "expenditures", "objectives"),
  lang = "eng"
)
```

The result is an interactive table displayed in the RStudio Viewer or a browser. The most important columns are:

- `title`: the official table title;
- `id`: the table number to pass to a download function;
- `release_date`: the catalogue release date; and
- `lang`: the language searched.

When several keywords are supplied, **all** of them must occur in the title. This makes searches precise, but it can also produce no matches. If that happens, remove one keyword or use a broader term:

```{r search-broad}
statcan_search("expenditures", lang = "eng")
```

Search French titles by using `lang = "fra"`:

```{r search-fra}
statcan_search(c("dépenses", "fédérales"), lang = "fra")
```

The catalogue is cached for 24 hours in R's platform-appropriate user cache directory. `statcan_find()` caches the table metadata used for geography checks for seven days. These caches make repeated searches faster and avoid unnecessary requests to Statistics Canada. Set `refresh = TRUE` only when you specifically need fresh information:

```{r search-refresh}
statcan_find(
  "population in Alberta since 2021",
  lang = "eng",
  refresh = TRUE
)
```

If WDS is temporarily unavailable, catalogue searches use the most recent valid cache and issue a warning. If geography metadata cannot be refreshed, `statcan_find()` uses valid cached metadata where possible. Candidates whose geography could not be checked have `geography_match = NA`; the match explanation makes that uncertainty explicit.

### Optional: ask a language model for help

`statcan_chat()` is an entirely optional layer on top of `statcan_find()`. It sends your query and `statcan_find()`'s ranked candidates to a language model you configure, which explains which candidate best matches and asks a clarifying question when the query is ambiguous. The language model never sees or chooses a table number of its own; it only explains the candidates `statcan_find()` already returned.

This feature adds no new package dependencies, and it never makes a network request unless you call `statcan_chat()` yourself. Two providers ship built in, selected with the `provider` argument: `"openai"` (the default) and `"anthropic"` (Claude). The endpoint defaults to the chosen provider and must use `https://` so the key is never sent in cleartext (plain `http://` is accepted only for a loopback host such as `http://localhost`, for a local model). Supply the API key through the provider's environment variable (`OPENAI_API_KEY` or `ANTHROPIC_API_KEY`, or the generic `STATCANR_LLM_API_KEY`, or the `api_key` argument), since a secret should not be kept in `options()`:

```{r chat-configure}
Sys.setenv(OPENAI_API_KEY = "sk-...")        # OpenAI
Sys.setenv(ANTHROPIC_API_KEY = "sk-ant-...") # Anthropic (Claude)
```

Loading the package with `library(statcanR)` never requests these settings, and no key is needed for `statcan_find()` or `statcan_data()`; the key is read only when you call `statcan_chat()`. To set it once and reuse it across sessions without retyping it, add it to your `~/.Renviron` file (open it with `usethis::edit_r_environ()`, then restart R) instead of calling `Sys.setenv()` each time:

```
OPENAI_API_KEY=sk-...
ANTHROPIC_API_KEY=sk-ant-...
```

R reads `.Renviron` automatically at startup, so the key stays out of your scripts and `.Rhistory`.

```{r chat-use}
# OpenAI (the default provider)
statcan_chat("R&D expenditures in Quebec since 2020", model = "gpt-4o-mini")

# Anthropic (Claude)
statcan_chat(
  "R&D expenditures in Quebec since 2020",
  provider = "anthropic", model = "claude-opus-4-8"
)
```

The `"openai"` provider also covers any OpenAI-compatible server — Groq, Together, OpenRouter, Mistral, vLLM, or a local open-source model served by Ollama or LM Studio — by pointing `endpoint` at it.

A single `statcan_chat()` call is one turn. To answer its `clarifying_question` and keep the discussion going, use `statcan_chat_continue()`. Each follow-up stays scoped to the same candidate tables the first call found (it never re-runs `statcan_find()`, and the model still never invents a table number), and results are chainable. Reassign the same variable each turn — there is no need for `r1`/`r2`/`r3`-style names:

```{r chat-continue}
chat <- statcan_chat("R&D spending in Quebec", provider = "anthropic", model = "claude-opus-4-8")
chat$clarifying_question
chat <- statcan_chat_continue(chat, "annual data, since 2015")
chat <- statcan_chat_continue(chat, "just the total, not by industry")
```

To search the catalogue again, start a fresh `statcan_chat()` call rather than continuing.

## Step 2: download a complete table

After choosing an identifier, pass it and the desired language to `statcan_data()`. This example uses a relatively small table that is convenient for learning:

```{r data-eng}
table_data <- statcan_data("10-10-0001-01", lang = "eng")
```

The function downloads and unpacks the current full-table CSV archive, then returns a data frame. Start by examining its dimensions, names, and first observations:

```{r inspect}
dim(table_data)
names(table_data)
head(table_data)
```

The exact columns depend on the table selected. `statcanR` applies a few consistent rules:

- the first column is named `REF_DATE`;
- annual, monthly, daily, and fiscal reference periods are converted to `Date` values when they can be interpreted safely;
- coordinate columns remain character values so leading zeros and compound coordinates are not lost; and
- `INDICATOR` contains the official table title read from its metadata.

For example, you can select observations from 2020 onward with ordinary R subsetting:

```{r subset}
recent_data <- table_data[
  !is.na(table_data$REF_DATE) &
    table_data$REF_DATE >= as.Date("2020-01-01"),
]
```

To download the French version of the table, change the language:

```{r data-fra}
table_fr <- statcan_data("10-10-0001-01", lang = "fra")
```

Most source column names remain in the selected language, so do not assume that every English column name has an identical French equivalent.

## Step 3: save a CSV copy when needed

If you only need to analyse the data in the current R session, `statcan_data()` is sufficient. Use `statcan_download_data()` when you also need a CSV file.

Existing two-argument calls save the file in the current working directory:

```{r save-default}
table_data <- statcan_download_data("10-10-0001-01", "eng")
getwd()
```

This creates `statcan_10100001_eng.csv`. To keep project files organized, create a dedicated directory and pass it through `path`:

```{r save-path}
output_dir <- file.path(tempdir(), "statcanR-data")
dir.create(output_dir, recursive = TRUE, showWarnings = FALSE)

table_data <- statcan_download_data(
  "10-10-0001-01",
  "eng",
  path = output_dir
)

attr(table_data, "statcan_file")
```

The output directory must already exist. The function returns the same data frame as `statcan_data()` and stores the exact CSV path in its `statcan_file` attribute. The CSV uses UTF-8 encoding, excludes R row names, and writes missing values as empty fields.

## Compatibility with earlier statcanR scripts

The update does not require you to rewrite established calls:

```{r compatibility}
# This familiar two-argument form remains valid.
table_data <- statcan_data("10-10-0001-01", "eng")

# This also remains valid and saves into the working directory.
table_data <- statcan_download_data("10-10-0001-01", "eng")
```

The optional `path` argument extends `statcan_download_data()` without changing the meaning of the original arguments. Both hyphenated table numbers and eight-digit PIDs are accepted.

## Troubleshooting

The package validates inputs before downloading and reports network or service problems explicitly. Common issues include:

| Message or symptom | What to check |
|---|---|
| No natural-language results | Keep a clear subject, but remove a geography or date constraint; then inspect broader candidates |
| No exact keyword results | Try fewer keywords, check the selected language, or use a broader official term |
| Invalid `tableNumber` | Use a displayed number such as `10-10-0001-01` or an eight-digit PID such as `10100001` |
| Invalid `lang` | Use exactly `"eng"` or `"fra"` |
| Output directory does not exist | Create the directory before supplying it through `path` |
| WDS is unavailable | Check the internet connection and try again later; catalogue search may use a valid cache |
| Download takes a long time | The function retrieves the complete table, which may be large |

Network failures, invalid tables, unexpected API responses, and malformed archives stop with messages that identify the affected Product ID. Temporary files created by a call are removed when it finishes; other files in the R session's temporary directory are left untouched.

## Reproducible use

WDS provides the current version of a Statistics Canada table, and published observations may be revised. For work that must be reproduced later:

1. record the table identifier, language, package version, and retrieval date;
2. save a local CSV copy of the data used in the analysis; and
3. cite the table and Statistics Canada according to the applicable data licence.

## Data licence and citation

Review the [Statistics Canada Open Licence](https://www.statcan.gc.ca/en/terms-conditions/open-licence) before redistributing downloaded data. To obtain the package's current citation, run:

```{r citation}
citation("statcanR")
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/statcan_chat.R
\name{statcan_chat_continue}
\alias{statcan_chat_continue}
\title{Continue a \code{statcan_chat()} conversation}
\usage{
statcan_chat_continue(result, message, api_key = NULL)
}
\arguments{
\item{result}{A \code{statcan_chat_result} returned by \code{\link[=statcan_chat]{statcan_chat()}} (or by a
previous \code{statcan_chat_continue()} call).}

\item{message}{One non-empty character string: your follow-up to the model.}

\item{api_key}{API key, re-resolved exactly as in \code{\link[=statcan_chat]{statcan_chat()}}. Defaults
to \code{Sys.getenv("STATCANR_LLM_API_KEY")}, then the provider's native
variable. Not read from \code{options()}.}
}
\value{
An updated \code{statcan_chat_result} with the same \code{candidates}, a new
\code{explanation} and \code{clarifying_question}, and an extended conversation.
}
\description{
Sends a follow-up \code{message} -- typically an answer to the
\code{clarifying_question} from a previous turn -- and returns an updated
result. The follow-up stays scoped to the \strong{same candidate tables} that
\code{\link[=statcan_chat]{statcan_chat()}} already found: it never re-runs \code{\link[=statcan_find]{statcan_find()}} and the
model still never proposes a table number of its own. To search the
catalogue again, start a new conversation with \code{\link[=statcan_chat]{statcan_chat()}}.
}
\details{
The returned object is itself continuable, so you can chain several
follow-ups. The API key is re-resolved on each call (from the \code{api_key}
argument, the \code{STATCANR_LLM_API_KEY} environment variable, or the
provider's native variable) because, for safety, it is never stored in the
result object.
}
\examples{
\dontrun{
Sys.setenv(ANTHROPIC_API_KEY = "sk-ant-...")
chat <- statcan_chat(
  "R&D spending in Quebec",
  provider = "anthropic", model = "claude-opus-4-8"
)
chat$clarifying_question

# Answer it and keep the same shortlist of candidates. Reassigning the same
# variable keeps the latest turn; you never need r1/r2/r3-style names.
chat <- statcan_chat_continue(chat, "annual data, since 2015")
chat$explanation

# Chain another follow-up the same way:
chat <- statcan_chat_continue(chat, "just the total, not by industry")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/statCanR.R
\name{statcan_data}
\alias{statcan_data}
\title{Download a Statistics Canada data table}
\usage{
statcan_data(tableNumber, lang)
}
\arguments{
\item{tableNumber}{A Statistics Canada table number or Product ID. Both
\code{"27-10-0014-01"} and \code{"27100014"} are accepted.}

\item{lang}{Language of the downloaded table: \code{"eng"} or \code{"fra"}.}
}
\value{
A data frame containing the complete Statistics Canada table.
}
\description{
Downloads a complete table from Statistics Canada's Web Data Service (WDS)
and returns it as a data frame. Product IDs can be supplied in the familiar
hyphenated form (for example, \code{"27-10-0014-01"}) or as an eight-digit PID
(for example, \code{"27100014"}).
}
\details{
The function keeps the interface used by earlier statcanR releases. English
and French downloads share the same processing rules. In particular, the
first column is named \code{REF_DATE}, coordinates are stored as character, and
the table title from the metadata file is added as \code{INDICATOR}.
}
\examples{
\dontrun{
science <- statcan_data("27-10-0014-01", "eng")
science_fr <- statcan_data("27100014", "fra")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/statcan_find.R
\name{statcan_find}
\alias{statcan_find}
\title{Find Statistics Canada tables using a natural-language query}
\usage{
statcan_find(query, lang = c("eng", "fra"), n = 5L, refresh = FALSE)
}
\arguments{
\item{query}{One non-empty character string describing the desired data.}

\item{lang}{Language of the table titles: \code{"eng"} or \code{"fra"}.}

\item{n}{Maximum number of candidates to return, from 1 to 20.}

\item{refresh}{Logical; if \code{TRUE}, request a fresh catalogue and fresh
candidate metadata from WDS.}
}
\value{
A data frame of ranked candidates. \code{id} contains the table number,
\code{score} is the relevance score, and \code{match_reason} explains the ranking.
Coverage dates describe the table as a whole. \code{geography_match} is \code{TRUE}
when WDS metadata confirms every geography in the query, \code{FALSE} when it
does not, and \code{NA} when no geography was requested or validation was not
possible.
}
\description{
Interprets a short description of the data needed and returns ranked
Statistics Canada table candidates. The query can contain a topic, Canadian
geography, and reference year or range. For example,
\code{"R&D expenditures in Quebec since 2020"} is interpreted as a research and
development expenditure topic, a Quebec geography constraint, and coverage
beginning in 2020.
}
\details{
This function finds tables; it does not download or filter their
observations. Use the returned \code{id} with \code{\link[=statcan_data]{statcan_data()}}, then apply any
required geography and date filters to the downloaded data. Rankings are a
discovery aid, so review the candidate title before downloading a large
table.

The catalogue is cached for 24 hours. When a geography is present in the
query, metadata for a small set of leading candidates is retrieved through
WDS to confirm that the geography is a table member. That metadata is cached
for seven days. If metadata is temporarily unavailable, candidates can
still be returned with \code{geography_match = NA}.
}
\examples{
\dontrun{
matches <- statcan_find(
  "R&D expenditures in Quebec since 2020",
  lang = "eng"
)
matches[, c("title", "id", "match_reason")]
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/statcan_search.R
\name{statcan_search}
\alias{statcan_search}
\title{Search Statistics Canada data tables}
\usage{
statcan_search(keywords, lang = c("eng", "fra"), refresh = FALSE)
}
\arguments{
\item{keywords}{Character vector of words that must appear in the table
title. Matching is case-insensitive.}

\item{lang}{Language of the returned titles: \code{"eng"} or \code{"fra"}.}

\item{refresh}{Logical; if \code{TRUE}, ignore a fresh cache and request the
catalogue from WDS.}
}
\value{
A \code{DT::datatable} containing matching table titles, Product IDs,
release dates, and language.
}
\description{
Searches the catalogue of tables published through Statistics Canada's Web
Data Service (WDS). The catalogue is cached for 24 hours in the user's cache
directory. If WDS is temporarily unavailable, the most recent valid cache is
used.
}
\examples{
\dontrun{
statcan_search(c("economy", "export"), "eng")
statcan_search("population", "fra")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/statCanR.R
\name{statcan_download_data}
\alias{statcan_download_data}
\title{Download a Statistics Canada table and save a CSV file}
\usage{
statcan_download_data(tableNumber, lang, path = ".")
}
\arguments{
\item{tableNumber}{A Statistics Canada table number or Product ID. Both
\code{"27-10-0014-01"} and \code{"27100014"} are accepted.}

\item{lang}{Language of the downloaded table: \code{"eng"} or \code{"fra"}.}

\item{path}{Directory in which to save the CSV file. The directory must
already exist. Defaults to the current working directory.}
}
\value{
The downloaded table. The CSV path is available in the
\code{statcan_file} attribute of the returned data frame.
}
\description{
Calls \code{\link[=statcan_data]{statcan_data()}} and also writes the returned table to a CSV file.
Existing calls with only \code{tableNumber} and \code{lang} remain supported; \code{path}
can be used to select a different output directory.
}
\examples{
\dontrun{
science <- statcan_download_data("27-10-0014-01", "eng")
science <- statcan_download_data(
  "27-10-0014-01", "eng", path = tempdir()
)
attr(science, "statcan_file")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/statcan_chat.R
\name{statcan_chat}
\alias{statcan_chat}
\title{Get an LLM's help interpreting a natural-language table search}
\usage{
statcan_chat(
  query,
  lang = c("eng", "fra"),
  n = 5L,
  refresh = FALSE,
  endpoint = NULL,
  api_key = NULL,
  model = NULL,
  provider = c("openai", "anthropic")
)
}
\arguments{
\item{query}{One non-empty character string describing the desired data.
Passed to \code{\link[=statcan_find]{statcan_find()}}.}

\item{lang}{Language of the table titles and of the model's reply:
\code{"eng"} or \code{"fra"}.}

\item{n}{Maximum number of candidates to request from \code{\link[=statcan_find]{statcan_find()}}.}

\item{refresh}{Logical; forwarded to \code{\link[=statcan_find]{statcan_find()}}.}

\item{endpoint}{Provider endpoint URL. Defaults to
\code{getOption("statcanR.llm_endpoint")}, then
\code{Sys.getenv("STATCANR_LLM_ENDPOINT")}, then the chosen provider's own
endpoint. Must be \verb{https://}, except for loopback hosts (for example,
\verb{http://localhost}).}

\item{api_key}{API key. Defaults to \code{Sys.getenv("STATCANR_LLM_API_KEY")},
then the provider's native variable (\code{OPENAI_API_KEY} for \code{"openai"},
\code{ANTHROPIC_API_KEY} for \code{"anthropic"}). For safety it is not read from
\code{options()}. It is sent as an \code{Authorization: Bearer} header for
\code{"openai"} and as an \code{x-api-key} header for \code{"anthropic"}.}

\item{model}{Model name sent to the provider (for example, \code{"gpt-4o-mini"}
for OpenAI or \code{"claude-opus-4-8"} for Anthropic). Defaults to
\code{getOption("statcanR.llm_model")}, then
\code{Sys.getenv("STATCANR_LLM_MODEL")}.}

\item{provider}{Which LLM provider to use: \code{"openai"} (the default,
also covering OpenAI-compatible and local servers) or \code{"anthropic"}
(Claude).}
}
\value{
A \code{statcan_chat_result} object: a list with \code{query}, \code{candidates}
(the \code{\link[=statcan_find]{statcan_find()}} data frame), \code{explanation}, and
\code{clarifying_question} (\code{NA} when the model had none).
}
\description{
Sends the query and the ranked candidates from \code{\link[=statcan_find]{statcan_find()}} to a
user-configured language-model provider, which explains which
candidate(s) best match and asks a clarifying question when the query is
ambiguous. The candidate table numbers and rankings always come from
\code{\link[=statcan_find]{statcan_find()}} itself; the language model only interprets and explains
them, and is never allowed to propose a table number of its own.
}
\details{
Two providers ship built in, selected with the \code{provider} argument:
\itemize{
\item \code{"openai"} (the default): the OpenAI chat-completions format, using an
\code{Authorization: Bearer} API key. This also covers any OpenAI-compatible
server -- Groq, Together, OpenRouter, Mistral, vLLM, or a local
open-source model served by Ollama or LM Studio -- by pointing
\code{endpoint} at it (a loopback \verb{http://localhost} endpoint is accepted for
local models; pass any placeholder \code{api_key} for servers that ignore
it).
\item \code{"anthropic"}: the Claude Messages format, using an \code{x-api-key} header.
}

This is an optional feature. It requires no additional packages beyond
what statcanR already imports, but it does require you to configure an
API key and model (the endpoint defaults to the chosen provider), either
as arguments or through \code{options()} / environment variables:
\itemize{
\item \code{endpoint}: \code{options(statcanR.llm_endpoint = ...)} or
\code{Sys.setenv(STATCANR_LLM_ENDPOINT = ...)}. Defaults to the provider's
own endpoint when unset.
\item \code{api_key}: \code{Sys.setenv(STATCANR_LLM_API_KEY = ...)} (or the \code{api_key}
argument, or the provider's native variable -- \code{OPENAI_API_KEY} /
\code{ANTHROPIC_API_KEY}). Because it is a secret, the key is \strong{not} read
from \code{options()}, which can be dumped, saved with a session, or
recorded in \code{.Rhistory}.
\item \code{model}: \code{options(statcanR.llm_model = ...)} or
\code{Sys.setenv(STATCANR_LLM_MODEL = ...)}
}

The endpoint must use \verb{https://} so the key is never sent in cleartext;
plain \verb{http://} is accepted only for loopback hosts (for example,
\verb{http://localhost} for a local model).

No network request is made unless \code{statcan_chat()} is called directly.
}
\examples{
\dontrun{
# OpenAI (the default provider)
Sys.setenv(OPENAI_API_KEY = "sk-...")
result <- statcan_chat(
  "R&D expenditures in Quebec since 2020",
  model = "gpt-4o-mini"
)

# Anthropic (Claude)
Sys.setenv(ANTHROPIC_API_KEY = "sk-ant-...")
result <- statcan_chat(
  "R&D expenditures in Quebec since 2020",
  provider = "anthropic", model = "claude-opus-4-8"
)

# A local open-source model served by Ollama (no key over loopback http)
result <- statcan_chat(
  "R&D expenditures in Quebec since 2020",
  endpoint = "http://localhost:11434/v1/chat/completions",
  api_key = "ollama", model = "llama3.1"
)

# statcan_chat() returns several ranked candidates, not a single table:
# the model explains them but never picks or invents one for you.
# The candidates are already a data frame, so you never retype an id.
result$candidates          # the full statcan_find() data frame
result$candidates$id       # every candidate id, best-ranked first
result$candidates$id[1]    # just the top-ranked id
result$explanation         # the model's plain-language explanation

# Feed the chosen id straight into statcan_data() -- nothing copied by hand.
table_data <- statcan_data(result$candidates$id[1], lang = "eng")
}
}
